import numpy as np
import pytest
from lgdo import Array, ArrayOfEqualSizedArrays, Table, WaveformTable

from pygama.pargen import pz_correct


def test_tp100_align():
    """
    Generate synthetic waveforms with random tp100s, then ensure that this function aligns all tp100s at the same sample
    """
    np.random.seed(42)
    tp100_known = np.random.randint(3000, 3200, 1000)

    tau = 30000
    wf_len = 8192

    test_wfs = []
    for i in range(len(tp100_known)):
        xs = np.arange(0, wf_len - tp100_known[i])
        ys = np.exp(-xs / tau)
        test_wfs.append(np.insert(ys, 0, np.zeros(tp100_known[i])))

    superpulse_window_width = 13

    time_aligned_wfs = pz_correct.tp100_align(
        test_wfs, superpulse_window_width, tp100_known
    )
    tp100s = []
    for wf in time_aligned_wfs:
        tp100s.append(np.argmax(wf))
    assert np.array_equal(
        np.array(tp100s), np.full(len(tp100s), np.median(tp100_known) - 13)
    )
    assert len(time_aligned_wfs[0]) == 8192 - 2 * 13


def test_dpz_model_fit():
    tau1 = 30000
    tau2 = 1100
    frac = 0.02
    wf_len = 8192
    tp0 = 3200

    xs = np.arange(0, wf_len - tp0)
    ys = pz_correct.dpz_model(xs, 1000, tau1, tau2, frac)
    test_wf = np.insert(ys, 0, np.zeros(tp0))
    tau1_fit, tau2_fit, frac_fit, plot_dict_out = pz_correct.dpz_model_fit(
        test_wf, percent_tau1_fit=0.1, percent_tau2_fit=0.2, idx_shift=2, plot=0
    )

    assert np.allclose(tau1, tau1_fit, rtol=1e-2)
    assert np.allclose(tau2, tau2_fit, rtol=1e-2)
    assert np.allclose(frac, frac_fit, rtol=1e-2)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_get_dpz_decay_constants():
    """
    First, generate a fake HPGe energy spectrum and associated waveforms. Then extract the time constants from the waveforms associated with the 2615 peak
    """
    np.random.seed(42)
    tau1 = 30000
    tau2 = 1100
    frac = 0.02
    wf_len = 8192
    tp0 = 3200
    num_wfs = 1000

    # Make the waveforms and then the lgdo table
    # peaks = np.array([238.632, 583.191, 727.330, 860.564, 1620.5, 2103.53, 2614.553])
    # daq_energies = []
    # for peak in peaks:
    #     daq_energies.extend(np.random.normal(peak, 10, num_wfs_per_peak).astype(int))

    daq_energies = np.random.normal(2614.553, 10, num_wfs).astype(int)

    wfs = []
    for amplitude in daq_energies:
        xs = np.arange(0, wf_len - tp0)
        ys = pz_correct.dpz_model(xs, amplitude, tau1, tau2, frac)
        test_wf = np.insert(ys, 0, np.zeros(tp0))
        wfs.append(test_wf)

    wf_tb = WaveformTable(
        dt=np.full(len(wfs), 16),
        dt_units="ns",
        t0=np.full(len(wfs), 0),
        t0_units="ns",
        values=ArrayOfEqualSizedArrays(nda=np.array(wfs)),
        values_units="ADC",
    )
    tb_data = Table(size=len(wfs))
    tb_data.add_column(name="daqenergy", obj=Array(daq_energies))
    tb_data.add_column(name="waveform", obj=wf_tb)

    assert isinstance(tb_data["waveform"]["values"], ArrayOfEqualSizedArrays)

    dpz_opt_dsp_dict = dict(
        {
            "outputs": ["tau1", "tau2", "frac"],
            "processors": {
                "bl_mean , bl_std, bl_slope, bl_intercept": {
                    "function": "linear_slope_fit",
                    "module": "dspeed.processors",
                    "args": [
                        "waveform[0:200]",
                        "bl_mean",
                        "bl_std",
                        "bl_slope",
                        "bl_intercept",
                    ],
                    "unit": ["ADC", "ADC", "ADC", "ADC"],
                },
                "wf_bl": {
                    "function": "bl_subtract",
                    "module": "dspeed.processors",
                    "args": ["waveform", "bl_mean", "wf_bl"],
                    "unit": "ADC",
                },
                "tau1, tau2, frac": {
                    "function": "optimize_2pz",
                    "module": "dspeed.processors",
                    "args": [
                        "waveform",
                        "bl_mean",
                        "round(52*us/waveform.period)",
                        "len(waveform)",
                        "1*ms/waveform.period",
                        "0.5",
                        "db.pz.tau1",
                        "db.pz.tau2",
                        "db.pz.frac",
                        "tau1",
                        "tau2",
                        "frac",
                    ],
                    "defaults": {
                        "db.pz.tau1": 28000,
                        "db.pz.tau2": 900,
                        "db.pz.frac": 0.145,
                    },
                },
            },
        }
    )

    wf_field = "waveform"
    percent_tau1_fit = 0.1
    percent_tau2_fit = 0.2
    offset_from_wf_max = 2
    superpulse_bl_idx = 25
    superpulse_window_width = 13
    display = 0

    tau = pz_correct.PZCorrect(dpz_opt_dsp_dict, wf_field)

    tau.get_dpz_decay_constants(
        tb_data,
        percent_tau1_fit,
        percent_tau2_fit,
        offset_from_wf_max,
        superpulse_bl_idx,
        superpulse_window_width,
        display,
    )
    assert np.allclose(
        float(tau.output_dict["pz"]["tau1"].split("*")[0]) / 16, tau1, rtol=1e-2
    )
    assert np.allclose(
        float(tau.output_dict["pz"]["tau2"].split("*")[0]) / 16, tau2, rtol=1e-2
    )
    assert np.allclose(float(tau.output_dict["pz"]["frac"]), frac, rtol=1e-2)



import numpy as np
from scipy.stats import expon, norm

from pygama.math.functions.gauss_on_exponential import gauss_on_exponential
from pygama.math.functions.sum_dists import SumDists


def test_gauss_on_exponential_pdf():
    x = np.arange(-10, 10)
    n_sig = 200
    mu = 1
    sigma = 1
    n_bkg = 100
    lamb = 3
    mu_exp = 2
    sigma_exp = 0.2

    pars = np.array([n_sig, mu, sigma, n_bkg, lamb, mu_exp, sigma_exp], dtype=float)

    assert isinstance(gauss_on_exponential, SumDists)

    y_direct = gauss_on_exponential.get_pdf(x, *pars)
    scipy_exponential = n_bkg * expon.pdf(x, mu_exp, sigma_exp / lamb)
    scipy_gauss = n_sig * norm.pdf(x, mu, sigma)

    scipy_y = scipy_exponential + scipy_gauss

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    x_lo = -100
    x_hi = 100

    pars = np.array(
        [x_lo, x_hi, n_sig, mu, sigma, n_bkg, lamb, mu_exp, sigma_exp], dtype=float
    )

    y_sig, y_ext = gauss_on_exponential.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig + n_bkg, rtol=1e-8)


def test_gauss_on_exponential_cdf():
    x = np.arange(-10, 10)
    n_sig = 200
    mu = 1
    sigma = 1
    n_bkg = 100
    lamb = 3
    mu_exp = 2
    sigma_exp = 0.2

    pars = np.array([n_sig, mu, sigma, n_bkg, lamb, mu_exp, sigma_exp], dtype=float)

    assert isinstance(gauss_on_exponential, SumDists)

    y_direct = gauss_on_exponential.get_cdf(x, *pars)

    scipy_exponential = n_bkg * expon.cdf(x, mu_exp, sigma_exp / lamb)
    scipy_gauss = n_sig * norm.cdf(x, mu, sigma)

    scipy_y = scipy_exponential + scipy_gauss

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = gauss_on_exponential.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_required_args():
    names = gauss_on_exponential.required_args()
    assert names[0] == "n_sig"
    assert names[1] == "mu"
    assert names[2] == "sigma"
    assert names[3] == "n_bkg"
    assert names[4] == "lambd"
    assert names[5] == "mu_exp"
    assert names[6] == "sigma_exp"


def test_name():
    assert gauss_on_exponential.name == "gauss_on_exponential"



r"""
Provide a convenience function for a Gaussian on top of a uniform distribution.

Parameters
----------
x_lo, x_hi
    The upper and lower bounds on which the uniform distribution is defined
n_sig
    The area of the Gaussian distribution
mu, sigma
    The location and scale of the first Gaussian
n_bkg
    The area of the uniform distributions respectively

Returns
-------
gauss_on_uniform
    An instance of SumDists and rv_continuous, has methods of pdf, cdf, etc.
"""

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.sum_dists import SumDists
from pygama.math.functions.uniform import uniform

(x_lo, x_hi, n_sig, mu, sigma, n_bkg) = range(
    6
)  # the sum_dist array we will pass will be x_lo, x_hi, n_sig, mu, sigma, n_bkg
par_array = [(gaussian, [mu, sigma]), (uniform, [x_lo, x_hi])]

gauss_on_uniform = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "n_bkg"],
    name="gauss_on_uniform",
)



"""
This module is for creating dplms dictionary for ge processing
"""

from __future__ import annotations

import itertools
import logging
import time

import matplotlib.pyplot as plt
import numpy as np
from lgdo import Table
from scipy.signal import convolve, convolve2d
from scipy.stats import chi2

import pygama.math.distributions as pmd  # noqa: pycln, F401
from pygama.pargen.data_cleaning import generate_cuts
from pygama.pargen.dsp_optimize import run_one_dsp
from pygama.pargen.energy_optimisation import fom_fwhm_with_alpha_fit

log = logging.getLogger(__name__)


def dplms_ge_dict(
    raw_fft: Table,
    raw_cal: Table,
    dsp_config: dict,
    par_dsp: dict,
    dplms_dict: dict,
    fom_func,
    decay_const: float = 0,
    ene_par: str = "dplmsEmax",
    display: int = 0,
) -> dict:
    """
    This function calculates the dplms dictionary for HPGe detectors.

    Parameters
    ----------
    raw_fft
        table with fft raw data
    raw_cal
        table with cal raw data
    dsp_config
        dsp config file
    par_dsp
        Dictionary with db parameters for dsp processing
    dplms_dict
        Dictionary with various parameters
    fom_func
        Function for peak fits

    Returns
    -------
    out_dict
    """

    t0 = time.time()
    log.info("Selecting baselines")

    dsp_fft = run_one_dsp(raw_fft, dsp_config, db_dict=par_dsp)

    cut_dict = generate_cuts(dsp_fft, cut_dict=dplms_dict["bls_cut_pars"])
    log.debug(f"Cuts are {cut_dict}")
    idxs = np.full(len(dsp_fft), True, dtype=bool)
    for outname, info in cut_dict.items():
        outcol = dsp_fft.eval(info["expression"], info.get("parameters", None))
        dsp_fft.add_column(outname, outcol)
    for cut in cut_dict:
        idxs = dsp_fft[cut].nda & idxs
    log.debug("Applied Cuts")

    bl_field = dplms_dict["bl_field"]
    log.info(f"... {len(dsp_fft[bl_field].values.nda[idxs, :])} baselines after cuts")

    bls = dsp_fft[bl_field].values.nda[idxs, : dplms_dict["bsize"]]
    bls_par = {}
    bls_cut_pars = [par for par in dplms_dict["bls_cut_pars"]]
    for par in bls_cut_pars:
        bls_par[par] = dsp_fft[dplms_dict["bls_cut_pars"][par]["cut_parameter"]].nda
    t1 = time.time()
    log.info(
        f"total events {len(raw_fft)}, {len(bls)} baseline selected in {(t1-t0):.2f} s"
    )

    log.info(
        f'Calculating noise matrix of length {dplms_dict["length"]} n. events: {bls.shape[0]}, size: {bls.shape[1]}'
    )
    nmat = noise_matrix(bls, dplms_dict["length"])
    t2 = time.time()
    log.info(f"Time to calculate noise matrix {(t2-t1):.2f} s")

    log.info("Selecting signals")
    wsize = dplms_dict["wsize"]
    wf_field = dplms_dict["wf_field"]
    peaks_kev = np.array(dplms_dict["peaks_kev"])
    kev_widths = [tuple(kev_width) for kev_width in dplms_dict["kev_widths"]]

    log.info(f"Produce dsp data for {len(raw_cal)} events")
    dsp_cal = run_one_dsp(raw_cal, dsp_config, db_dict=par_dsp)
    t3 = time.time()
    log.info(f"Time to run dsp production {(t3-t2):.2f} s")

    dsp_config["outputs"] = [ene_par, "dt_eff"]

    # dictionary for peak fitting
    peak_dict = {
        "peak": peaks_kev[-1],
        "kev_width": kev_widths[-1],
        "parameter": ene_par,
        "func": fom_func,
    }

    if display > 0:
        plot_dict = {}
        plot_dict["dplms"] = {}

    # penalized coefficients
    dp_coeffs = dplms_dict["dp_coeffs"]
    za_coeff = dplms_dict["dp_def"]["za"]
    dp_coeffs.pop("za")
    coeff_keys = [key for key in dp_coeffs.keys()]
    lists = [dp_coeffs[key] for key in dp_coeffs.keys()]

    prod = list(itertools.product(*lists))
    grid_dict = {}
    min_fom = float("inf")
    min_idx = None

    for i, values in enumerate(prod):
        coeff_values = dict(zip(coeff_keys, values))
        log_msg = f"Case {i} ->"
        for key, value in coeff_values.items():
            log_msg += f" {key} = {value}"
        log.info(log_msg)

        grid_dict[i] = coeff_values

        sel_dict = signal_selection(dsp_cal, dplms_dict, coeff_values)
        wfs = dsp_cal[wf_field].nda[sel_dict["idxs"], :]
        log.info(f"... {len(wfs)} signals after signal selection")

        ref, rmat, pmat, fmat = signal_matrices(wfs, dplms_dict["length"], decay_const)

        t_tmp = time.time()
        nm_coeff = coeff_values["nm"]
        ft_coeff = coeff_values["ft"]
        x, y, refy = filter_synthesis(
            ref,
            nm_coeff * nmat,
            rmat,
            za_coeff,
            pmat,
            ft_coeff * fmat,
            dplms_dict["length"],
            wsize,
        )
        par_dsp["dplms"] = {"length": dplms_dict["length"], "coefficients": x}
        log.info(
            f"Filter synthesis in {time.time()-t_tmp:.1f} s, filter area {np.sum(x)}"
        )

        t_tmp = time.time()
        dsp_opt = run_one_dsp(raw_cal, dsp_config, db_dict=par_dsp)

        try:
            res = fom_fwhm_with_alpha_fit(
                dsp_opt,
                peak_dict,
                "dt_eff",
                idxs=np.where(~np.isnan(dsp_opt["dt_eff"].nda))[0],
                frac_max=0.5,
            )
        except Exception:
            log.debug("FWHM not calculated")
            continue

        fwhm, fwhm_err, alpha, chisquare = (
            res["fwhm"],
            res["fwhm_err"],
            res["alpha"],
            res["chisquare"],
        )
        p_val = chi2.sf(chisquare[0], chisquare[1])
        log.info(
            f"FWHM = {fwhm:.2f} ± {fwhm_err:.2f} keV, p_val={p_val} evaluated in {time.time()-t_tmp:.1f} s"
        )
        grid_dict[i]["fwhm"] = fwhm
        grid_dict[i]["fwhm_err"] = fwhm_err
        grid_dict[i]["alpha"] = alpha
        if (
            fwhm < dplms_dict["fwhm_limit"]
            and fwhm_err < dplms_dict["err_limit"]
            and p_val > dplms_dict["p_val_lim"]
            and ~np.isnan(fwhm)
        ):
            if fwhm < min_fom:
                min_idx, min_fom = i, fwhm

    if min_idx is not None:
        min_result = grid_dict[min_idx]
        best_case_values = {key: min_result[key] for key in min_result.keys()}

        fwhm = best_case_values.get("fwhm", None)
        fwhm_err = best_case_values.get("fwhm_err", 0)
        alpha = best_case_values.get("alpha", 0)
        nm_coeff = best_case_values.get("nm", dplms_dict["dp_def"]["nm"])
        ft_coeff = best_case_values.get("ft", dplms_dict["dp_def"]["nm"])
        rt_coeff = best_case_values.get("rt", dplms_dict["dp_def"]["rt"])
        pt_coeff = best_case_values.get("pt", dplms_dict["dp_def"]["pt"])
        if all(
            v is not None
            for v in [
                fwhm,
                fwhm_err,
                alpha,
                nm_coeff,
                ft_coeff,
                rt_coeff,
                pt_coeff,
            ]
        ):
            log.info(
                f"\nBest case: FWHM = {fwhm:.2f} ± {fwhm_err:.2f} keV, ctc {alpha}"
            )
        else:
            log.debug("Some values are missing in the best case results")
    else:
        log.debug("Filter synthesis failed")
        nm_coeff = dplms_dict["dp_def"]["nm"]
        ft_coeff = dplms_dict["dp_def"]["ft"]
        rt_coeff = dplms_dict["dp_def"]["rt"]
        pt_coeff = dplms_dict["dp_def"]["pt"]
        best_case_values = {}
        alpha = 0

    # filter synthesis
    sel_dict = signal_selection(dsp_cal, dplms_dict, best_case_values)
    idxs = sel_dict["idxs"]
    wfs = dsp_cal[wf_field].nda[idxs, :]
    ref, rmat, pmat, fmat = signal_matrices(wfs, dplms_dict["length"], decay_const)

    x, y, refy = filter_synthesis(
        ref,
        nm_coeff * nmat,
        rmat,
        za_coeff,
        pmat,
        ft_coeff * fmat,
        dplms_dict["length"],
        wsize,
    )

    out_dict = {
        "dplms": {
            "length": dplms_dict["length"],
            "coefficients": x,
            "dp_coeffs": {
                "nm": nm_coeff,
                "za": za_coeff,
                "ft": ft_coeff,
                "rt": rt_coeff,
                "pt": pt_coeff,
            },
        }
    }
    out_alpha_dict = {
        f"{ene_par}_ctc": {
            "expression": f"{ene_par}*(1+dt_eff*a)",
            "parameters": {"a": round(alpha, 9)},
        }
    }
    out_dict.update({"ctc_params": out_alpha_dict})

    log.info(f"Time to complete DPLMS filter synthesis {time.time()-t0:.1f}")

    if display > 0:
        plot_dict = {"ref": ref, "coefficients": x}

        bl_idxs = np.random.choice(len(bls), dplms_dict["n_plot"])
        bls = bls[bl_idxs]
        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        for ii, wf in enumerate(bls):
            if ii < 10:
                ax.plot(wf, label=f"mean = {wf.mean():.1f}")
            else:
                ax.plot(wf)
        ax.legend(loc="upper right")
        plot_dict["bls"] = fig
        fig, ax = plt.subplots(nrows=2, ncols=3, figsize=(16, 9), facecolor="white")
        for ii, par in enumerate(bls_cut_pars):
            if "parameters" in cut_dict[par]:
                if "a" in cut_dict[par]["parameters"]:
                    llo = cut_dict[par]["parameters"]["a"]
                else:
                    llo = np.nan
                if "b" in cut_dict[par]["parameters"]:
                    lup = cut_dict[par]["parameters"]["b"]
                else:
                    lup = np.nan
            mean = (lup + llo) / 2
            plo, pup = mean - 2 * (mean - llo), mean + 2 * (lup - mean)
            hh, bb = np.histogram(bls_par[par], bins=np.linspace(plo, pup, 200))
            ax.flat[ii].plot(bb[1:], hh, ds="steps", label=f"cut on {par}")
            ax.flat[ii].axvline(lup, color="k", linestyle=":", label="selection")
            ax.flat[ii].axvline(llo, color="k", linestyle=":")
            ax.flat[ii].set_xlabel(par)
            ax.flat[ii].set_yscale("log")
            ax.flat[ii].legend(loc="upper right")
        plot_dict["bl_sel"] = fig

        wf_idxs = np.random.choice(len(wfs), dplms_dict["n_plot"])
        wfs = wfs[wf_idxs]
        centroid = dsp_cal["centroid"].nda

        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        for ii, wf in enumerate(wfs):
            if ii < 10:
                ax.plot(wf, label=f"centr = {centroid[ii]}")
            else:
                ax.plot(wf)
        ax.legend(loc="upper right")
        axin = ax.inset_axes([0.1, 0.15, 0.35, 0.5])
        for wf in wfs:
            axin.plot(wf)
        axin.set_xlim(wsize / 2 - dplms_dict["zoom"], wsize / 2 + dplms_dict["zoom"])
        axin.set_yticklabels("")
        plot_dict["wfs"] = fig

        peak_pos = dsp_cal["peak_pos"].nda
        risetime = dsp_cal["tp_90"].nda - dsp_cal["tp_10"].nda
        rt_low = dplms_dict["rt_low"]
        rt_high = dplms_dict["rt_high"]
        peak_lim = dplms_dict["peak_lim"]
        cal_par = {}
        wfs_cut_pars = ["centroid", "peak_pos", "risetime"]

        fig, ax = plt.subplots(nrows=2, ncols=3, figsize=(16, 9), facecolor="white")

        for ii, par in enumerate(wfs_cut_pars):
            pspace = np.linspace(
                wsize / 2 - peak_lim, wsize / 2 + peak_lim, 2 * peak_lim
            )
            if par == "centroid":
                llo, lup = sel_dict["ct_ll"], sel_dict["ct_hh"]
                hh, bb = np.histogram(centroid, bins=pspace)
            elif par == "peak_pos":
                llo, lup = sel_dict["pp_ll"], sel_dict["pp_hh"]
                hh, bb = np.histogram(peak_pos, bins=pspace)
            elif par == "risetime":
                llo, lup = sel_dict["rt_ll"], sel_dict["rt_hh"]
                rt_bins = int((rt_high - rt_low) / dplms_dict["period"])
                rt_space = np.linspace(rt_low, rt_high, rt_bins)
                hh, bb = np.histogram(risetime, bins=rt_space)
            else:
                llo, lup = np.min(cal_par[par]), np.max(cal_par[par])
                hh, bb = np.histogram(cal_par[par], bins=np.linspace(llo, lup, 200))
            ax.flat[ii + 1].plot(bb[1:], hh, ds="steps", label=f"cut on {par}")
            ax.flat[ii + 1].axvline(
                llo, color="k", linestyle=":", label=f"sel. {llo:.1f} {lup:.1f}"
            )
            if par != "centroid":
                ax.flat[ii + 1].axvline(lup, color="k", linestyle=":")
            ax.flat[ii + 1].set_xlabel(par)
            ax.flat[ii + 1].set_yscale("log")
            ax.flat[ii + 1].legend(loc="upper right")
        roughenergy = dsp_cal["trapTmax"].nda
        roughenergy_sel = roughenergy[idxs]
        ell, ehh = roughenergy.min(), roughenergy.max()
        he, be = np.histogram(roughenergy, bins=np.linspace(ell, ehh, 1000))
        hs, be = np.histogram(roughenergy_sel, bins=np.linspace(ell, ehh, 1000))
        ax.flat[0].plot(be[1:], he, c="b", ds="steps", label="initial")
        ax.flat[0].plot(be[1:], hs, c="r", ds="steps", label="selected")
        ax.flat[0].set_xlabel("rough energy (ADC)")
        ax.flat[0].set_yscale("log")
        ax.flat[0].legend(loc="upper right")
        plot_dict["wf_sel"] = fig

        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        ax.plot(x, "r-", label="filter")
        ax.axhline(0, color="black", linestyle=":")
        ax.legend(loc="upper right")
        axin = ax.inset_axes([0.6, 0.1, 0.35, 0.33])
        axin.plot(x, "r-")
        axin.set_xlim(
            dplms_dict["length"] / 2 - dplms_dict["zoom"],
            dplms_dict["length"] / 2 + dplms_dict["zoom"],
        )
        axin.set_yticklabels("")
        ax.indicate_inset_zoom(axin)

        return out_dict, plot_dict
    else:
        return out_dict


def is_valid_centroid(
    centroid: np.array, lim: int, size: int, full_size: int
) -> list[bool]:
    llim = size / 2 - lim
    hlim = full_size - size / 2
    idxs = (centroid > llim) & (centroid < hlim)
    return idxs, llim, hlim


def is_not_pile_up(
    peak_pos: np.array, peak_pos_neg: np.array, thr: int, lim: int, size: int
) -> list[bool]:
    bin_edges = np.linspace(size / 2 - lim, size / 2 + lim, 2 * lim)
    hist, bin_edges = np.histogram(peak_pos, bins=bin_edges)

    thr = thr * hist.max() / 100
    low_thr_idxs = np.where(hist[: hist.argmax()] < thr)[0]
    upp_thr_idxs = np.where(hist[hist.argmax() :] < thr)[0]

    idx_low = low_thr_idxs[-1] if low_thr_idxs.size > 0 else 0
    idx_upp = (
        upp_thr_idxs[0] + hist.argmax() if upp_thr_idxs.size > 0 else len(hist) - 1
    )

    llow, lupp = bin_edges[idx_low], bin_edges[idx_upp]

    idxs = []
    for n, nn in zip(peak_pos, peak_pos_neg):
        condition1 = np.count_nonzero(n > 0) == 1
        condition2 = (
            np.count_nonzero((n > 0) & ((n < llow) | (n > lupp) & (n < size))) == 0
        )
        condition3 = np.count_nonzero(nn > 0) == 0
        idxs.append(condition1 and condition2 and condition3)
    return idxs, llow, lupp


def is_valid_risetime(risetime: np.array, llim: int, perc: float):
    hlim = np.percentile(risetime[~np.isnan(risetime)], perc)
    idxs = (risetime >= llim) & (risetime <= hlim)
    return idxs, llim, hlim


def signal_selection(dsp_cal, dplms_dict, coeff_values):
    peak_pos = dsp_cal["peak_pos"].nda
    peak_pos_neg = dsp_cal["peak_pos_neg"].nda
    centroid = dsp_cal["centroid"].nda
    risetime = dsp_cal["tp_90"].nda - dsp_cal["tp_10"].nda

    rt_low = dplms_dict["rt_low"]
    peak_lim = dplms_dict["peak_lim"]
    wsize = dplms_dict["wsize"]
    bsize = dplms_dict["bsize"]

    centroid_lim = dplms_dict["centroid_lim"]
    if "rt" in coeff_values:
        perc = coeff_values["rt"]
    else:
        perc = dplms_dict["dp_def"]["rt"]
    if "pt" in coeff_values:
        thr = coeff_values["pt"]
    else:
        thr = dplms_dict["dp_def"]["rt"]

    idxs_ct, ct_ll, ct_hh = is_valid_centroid(centroid, centroid_lim, wsize, bsize)
    log.info(f"... {len(peak_pos[idxs_ct, :])} signals after alignment")

    idxs_pp, pp_ll, pp_hh = is_not_pile_up(peak_pos, peak_pos_neg, thr, peak_lim, wsize)
    log.info(f"... {len(peak_pos[idxs_pp, :])} signals after pile-up cut")

    idxs_rt, rt_ll, rt_hh = is_valid_risetime(risetime, rt_low, perc)
    log.info(f"... {len(peak_pos[idxs_rt, :])} signals after risetime cut")

    idxs = idxs_ct & idxs_pp & idxs_rt
    sel_dict = {
        "idxs": idxs,
        "ct_ll": ct_ll,
        "ct_hh": ct_hh,
        "pp_ll": pp_ll,
        "pp_hh": pp_hh,
        "rt_ll": rt_ll,
        "rt_hh": rt_hh,
    }
    return sel_dict


def noise_matrix(bls: np.array, length: int) -> np.array:
    nev, size = bls.shape
    ref = np.mean(bls, axis=0)
    offset = np.mean(ref)
    bls = bls - offset
    nmat = np.matmul(bls.T, bls, dtype=float) / nev
    kernel = np.identity(size - length + 1)
    nmat = convolve2d(nmat, kernel, boundary="symm", mode="valid") / (size - length + 1)
    return nmat


def signal_matrices(
    wfs: np.array, length: int, decay_const: float, ff: int = 2
) -> np.array:
    nev, size = wfs.shape
    lo = size // 2 - 100
    flo = size // 2 - length // 2
    fhi = size // 2 + length // 2
    offsets = np.mean(wfs[:, :lo], axis=1)
    wfs = wfs - offsets[:, np.newaxis]

    # Reference signal
    ref = np.sum(wfs, axis=0)
    ref /= np.max(ref)
    rmat = np.outer(ref[flo:fhi], ref[flo:fhi])

    # Pile-up matrix
    if decay_const > 0:
        decay = np.exp(-np.arange(length) / decay_const)
    else:
        decay = np.zeros(length)
    pmat = np.outer(decay, decay)

    # Flat top matrix
    flo -= ff // 2
    fhi += ff // 2
    wfs = wfs[:, flo:fhi]
    fmat = np.matmul(wfs.T, wfs, dtype=float) / nev
    m1 = ((1, -1), (-1, 1))
    fmat = convolve2d(fmat, m1, boundary="symm", mode="valid")
    if ff > 0:
        fmat = convolve2d(fmat, np.identity(ff), boundary="symm", mode="valid") / ff
    return ref, rmat, pmat, fmat


def filter_synthesis(
    ref: np.array,
    nmat: np.array,
    rmat: np.array,
    za: int,
    pmat: np.array,
    fmat: np.array,
    length: int,
    size: int,
    flip: bool = True,
) -> np.array:
    mat = nmat + rmat + za * np.ones([length, length]) + pmat + fmat
    flo = (size // 2) - (length // 2)
    fhi = (size // 2) + (length // 2)
    x = np.linalg.solve(mat, ref[flo:fhi]).astype(np.float32)
    y = convolve(ref, np.flip(x), mode="valid")
    maxy = np.max(y)
    x /= maxy
    y /= maxy
    refy = ref[(size // 2) - (len(y) // 2) : (size // 2) + (len(y) // 2)]
    if flip:
        return np.flip(x), y, refy
    else:
        return x, y, refy



import numpy as np
from scipy.stats import exponnorm, norm

from pygama.math.functions.hpge_peak import hpge_peak
from pygama.math.functions.sum_dists import SumDists


def test_hpge_peak_pdf():
    x = np.arange(-10, 10)

    sigma = 0.2
    mu = 2
    tau = 0.1
    htail = 0.75
    hstep = 0.5
    x_lo = np.amin(x)
    x_hi = np.max(x)
    n_sig = 10
    n_bkg = 20

    pars = np.array(
        [x_lo, x_hi, n_sig, mu, sigma, htail, tau, n_bkg, hstep],
        dtype=float,
    )

    assert isinstance(hpge_peak, SumDists)

    y_direct = hpge_peak.get_pdf(x, *pars)
    scipy_exgauss = htail * exponnorm.pdf(
        -1 * x, tau / sigma, -1 * mu, sigma
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma
    scipy_gauss = (1 - htail) * norm.pdf(x, mu, sigma)

    scipy_step = 1 + hstep * (norm.cdf(x, mu, sigma) * 2 - 1)

    maximum = (np.amax(x) - mu) / sigma
    minimum = (np.amin(x) - mu) / sigma
    normalization = sigma * (
        (
            maximum
            + hstep * maximum * (norm.cdf(maximum) * 2 - 1)
            + hstep * np.exp(-1 * maximum**2) / np.sqrt(np.pi)
        )
        - (
            minimum
            + hstep * minimum * (norm.cdf(minimum) * 2 - 1)
            + hstep * np.exp(-1 * minimum**2) / np.sqrt(np.pi)
        )
    )
    scipy_step = scipy_step / normalization

    scipy_y = n_sig * (scipy_exgauss + scipy_gauss) + n_bkg * scipy_step

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_sig, y_ext = hpge_peak.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig + n_bkg, rtol=1e-8)


def test_hpge_peak_cdf():
    x = np.arange(-10, 10)

    sigma = 0.2
    mu = 2
    tau = 0.1
    htail = 0.75
    hstep = 0.5
    x_lo = np.amin(x)
    x_hi = np.max(x)
    n_sig = 10
    n_bkg = 20

    pars = np.array(
        [x_lo, x_hi, n_sig, mu, sigma, htail, tau, n_bkg, hstep],
        dtype=float,
    )

    assert isinstance(hpge_peak, SumDists)

    y_direct = hpge_peak.get_cdf(x, *pars)

    scipy_exgauss = htail * (
        1 - exponnorm.cdf(-1 * x, tau / sigma, -1 * mu, sigma)
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma
    scipy_gauss = (1 - htail) * norm.cdf(x, mu, sigma)

    z = (x - mu) / sigma
    scipy_step = sigma * (
        z
        + hstep * z * (2 * norm.cdf(z, 0, 1) - 1)
        + hstep * np.exp(-1 * z**2) / np.sqrt(np.pi)
    )
    maximum = (np.amax(x) - mu) / sigma
    minimum = (np.amin(x) - mu) / sigma
    normalization = sigma * (
        (
            maximum
            + hstep * maximum * (norm.cdf(maximum) * 2 - 1)
            + hstep * np.exp(-1 * maximum**2) / np.sqrt(np.pi)
        )
        - (
            minimum
            + hstep * minimum * (norm.cdf(minimum) * 2 - 1)
            + hstep * np.exp(-1 * minimum**2) / np.sqrt(np.pi)
        )
    )

    scipy_step /= normalization
    scipy_hstep = scipy_step + (1 - scipy_step[-1])

    scipy_y = n_sig * (scipy_exgauss + scipy_gauss) + n_bkg * scipy_hstep

    assert np.allclose(y_direct, scipy_y, rtol=1e-1)

    y_ext = hpge_peak.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-1)


def test_required_args():
    names = hpge_peak.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "n_sig"
    assert names[3] == "mu"
    assert names[4] == "sigma"
    assert names[5] == "htail"
    assert names[6] == "tau"
    assert names[7] == "n_bkg"
    assert names[8] == "hstep"


def test_name():
    assert hpge_peak.name == "hpge_peak"



import numpy as np
from pytest import approx

import pygama.math.binned_fitting as pgbf


def test_fit_binned_and_goodness_of_fit():
    from numpy.random import normal

    from pygama.math.functions.gauss import nb_gauss_amp
    from pygama.math.histogram import get_hist

    np.random.seed(42)
    hist, bins, var = get_hist(normal(size=10000), bins=100, range=(-5, 5))
    fit, fit_error, fit_cov = pgbf.fit_binned(
        nb_gauss_amp, hist, bins, guess=(0, 0.9, 400), cost_func="Least Squares"
    )

    assert fit["mu"] == approx(-0.003933521046091256)
    assert fit["sigma"] == approx(0.9971419629418575)
    assert fit["a"] == approx(398.28437169188095)

    assert fit_error["mu"] == approx(0.010033194971234135)
    assert fit_error["sigma"] == approx(0.007291498082977738)
    assert fit_error["a"] == approx(4.930383892845187)

    assert np.allclose(
        fit_cov[0],
        np.array([[1.00665024e-04, -1.13519243e-06, 4.28011083e-04]], dtype=np.float32),
        rtol=1e-8,
    )

    chi, dof = pgbf.goodness_of_fit(
        hist, bins, var, nb_gauss_amp, fit, method="Pearson", scale_bins=True
    )
    assert chi == approx(82236.02419624529)
    assert dof == approx(97)

    poisson = pgbf.poisson_gof(fit, nb_gauss_amp, hist, bins, is_integral=False)

    assert poisson == approx(68204.56472918994)


def test_gauss_mode_width_max():
    from numpy.random import normal

    from pygama.math.histogram import get_hist

    np.random.seed(42)
    hist, bins, var = get_hist(normal(size=10000), bins=100, range=(-5, 5))
    fit, cov = pgbf.gauss_mode_width_max(hist, bins, n_bins=20)

    assert fit[0] == approx(-0.006127842485254326)
    assert fit[1] == approx(1.0066159284176235)
    assert fit[2] == approx(398.35078610804527)

    fit, err = pgbf.gauss_mode_max(hist, bins)

    assert fit[0] == approx(-0.0028635117051317638)
    assert fit[1] == approx(400.2130565573762)

    fit, err = pgbf.gauss_mode(hist, bins)

    assert fit == approx(-0.0028635117051317638)


def test_taylor_mode_max():
    from numpy.random import normal

    from pygama.math.histogram import get_hist

    np.random.seed(42)
    hist, bins, var = get_hist(normal(size=10000), bins=100, range=(-5, 5))
    fit, err = pgbf.taylor_mode_max(hist, bins, var=None, mode_guess=None, n_bins=5)

    assert fit[0] == approx(-0.005714285714285219)
    assert fit[1] == approx(400.4864285714285)



import numpy as np
from scipy.stats import poisson as scipy_poisson
from scipy.stats import rv_discrete

from pygama.math.functions.poisson import poisson


def test_poisson_pdf():
    x = np.arange(1, 12)
    mu = 1
    lamb = 2

    y = poisson.pmf(x, mu, lamb)
    y_direct = poisson.get_pmf(x, mu, lamb)
    scipy_y = scipy_poisson.pmf(x, lamb, mu)

    assert isinstance(poisson, rv_discrete)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = 0
    x_hi = 12

    y_sig, y_ext = poisson.pmf_ext(x, x_lo, x_hi, n_sig, mu, lamb)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-3)


def test_poisson_cdf():
    x = np.arange(0, 12)
    mu = 1
    lamb = 2

    y = poisson.cdf(x, mu, lamb)
    y_direct = poisson.get_cdf(x, mu, lamb)
    scipy_y = scipy_poisson.cdf(x, lamb, mu)

    assert isinstance(poisson, rv_discrete)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = poisson.cdf_ext(x, n_sig, mu, lamb)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)


def test_required_args():
    names = poisson.required_args()
    assert names[0] == "mu"
    assert names[1] == "lamb"


def test_name():
    assert poisson.name == "poisson"



import lgdo.lh5 as lh5
import numpy as np

import pygama.pargen.lq_cal as lq
from pygama.math.distributions import gaussian


def test_lq_cal(lgnd_test_data):
    # test the HPGe calibration function
    # the function should return a calibration polynomial
    # that maps ADC channel to energy in keV

    # load test data here
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )

    df = lh5.read_as("ch1104000/dsp", data, "pd")

    df["cuspEmax_cal"] = df["cuspEmax"] * 0.155

    cal_dict = {
        "LQ_Ecorr": {
            "expression": "lq80/cuspEmax",
            "parameters": {},
        }
    }

    lqcal = lq.LQCal(
        cal_dict,
        "cuspEmax_cal",
        "dt_eff",
        lambda x: np.sqrt(1.5 + 0.1 * x),
        selection_string="index==index",
        cdf=gaussian,
        debug_mode=True,
    )

    df["LQ_Ecorr"] = np.divide(df["lq80"], df["cuspEmax"])

    lqcal.calibrate(df, "LQ_Ecorr")
    assert (lqcal.cut_val > 0) & (~np.isnan(lqcal.cut_val))
    assert (~np.isnan(lqcal.low_side_sf.loc[1592.50]["sf"])) & (
        lqcal.low_side_sf.loc[1592.50]["sf"] > 95
    )



"""
This module provides aggregators to build the `evt` tier.
"""

from __future__ import annotations

import awkward as ak
import numpy as np
import pandas as pd
from lgdo import lh5, types

from . import utils


def evaluate_to_first_or_last(
    datainfo,
    tcm,
    channels,
    channels_skip,
    expr,
    field_list,
    query,
    n_rows,
    sorter,
    pars_dict=None,
    default_value=np.nan,
    is_first: bool = True,
    channel_mapping=None,
) -> types.Array:
    """Aggregates across channels by returning the expression of the channel
    with value of `sorter`.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    channels
       list of channels to be aggregated.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    query
       query expression to mask aggregation.
    n_rows
       length of output array.
    sorter
       tuple of field in `hit/dsp/evt` tier to evaluate ``(tier, field)``.
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    is_first
       defines if sorted by smallest or largest value of `sorter`
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    df = None

    for ch in channels:
        table_id = utils.get_tcm_id_by_pattern(datainfo.hit.table_fmt, ch)
        if table_id is None:
            continue

        # get index list for this channel to be loaded
        chan_tcm_indexs = ak.flatten(tcm.table_key) == table_id
        idx_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()

        # evaluate at channel
        if ch not in channels_skip:
            res = utils.get_data_at_channel(
                datainfo=datainfo,
                ch=ch,
                tcm=tcm,
                expr=expr,
                field_list=field_list,
                pars_dict=pars_dict,
            )

            if df is None:
                # define dimension of output array
                out = utils.make_numpy_full(n_rows, default_value, res.dtype)
                df = pd.DataFrame({"sort_field": np.zeros(len(out)), "res": out})

            # get mask from query
            limarr = utils.get_mask_from_query(
                datainfo=datainfo,
                query=query,
                length=len(res),
                ch=ch,
                idx_ch=idx_ch,
            )

            # find if sorter is in hit or dsp
            sort_field = lh5.read_as(
                f"{ch}/{sorter[0]}/{sorter[1]}",
                (
                    datainfo.hit.file
                    if f"{datainfo.hit.group}" == sorter[0]
                    else datainfo.dsp.file
                ),
                idx=idx_ch,
                library="np",
            )

            if sort_field.ndim > 1:
                raise ValueError(f"sorter '{sorter[0]}/{sorter[1]}' must be a 1D array")

            ch_df = pd.DataFrame({"sort_field": sort_field, "res": res})

            evt_ids_ch = np.repeat(
                np.arange(0, len(tcm.table_key)),
                ak.sum(tcm.table_key == table_id, axis=1),
            )

            if is_first:
                if ch == channels[0]:
                    df["sort_field"] = np.inf
                ids = (
                    ch_df.sort_field.to_numpy() < df.sort_field[evt_ids_ch].to_numpy()
                ) & (limarr)
            else:
                ids = (
                    ch_df.sort_field.to_numpy() > df.sort_field[evt_ids_ch].to_numpy()
                ) & (limarr)

            df.loc[evt_ids_ch[ids], list(df.columns)] = ch_df.loc[ids, list(df.columns)]

    return types.Array(nda=df.res.to_numpy())


def evaluate_to_scalar(
    datainfo,
    tcm,
    mode,
    channels,
    channels_skip,
    expr,
    field_list,
    query,
    n_rows,
    pars_dict=None,
    default_value=np.nan,
    channel_mapping=None,
) -> types.Array:
    """Aggregates by summation across channels.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    mode
       aggregation mode.
    channels
       list of channels to be aggregated.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    query
       query expression to mask aggregation.
    n_rows
       length of output array
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)
    out = None

    for ch in channels:
        table_id = utils.get_tcm_id_by_pattern(datainfo.hit.table_fmt, ch)
        if table_id is None:
            continue

        # get index list for this channel to be loaded
        chan_tcm_indexs = ak.flatten(tcm.table_key) == table_id
        idx_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()

        if ch not in channels_skip:
            res = utils.get_data_at_channel(
                datainfo=datainfo,
                ch=ch,
                tcm=tcm,
                expr=expr,
                field_list=field_list,
                pars_dict=pars_dict,
            )

            if out is None:
                # define dimension of output array
                out = utils.make_numpy_full(n_rows, default_value, res.dtype)

            # get mask from query
            limarr = utils.get_mask_from_query(
                datainfo=datainfo,
                query=query,
                length=len(res),
                ch=ch,
                idx_ch=idx_ch,
            )

            evt_ids_ch = np.repeat(
                np.arange(0, len(tcm.table_key)),
                ak.sum(tcm.table_key == table_id, axis=1),
            )

            # switch through modes
            if mode == "sum":
                if res.dtype == bool:
                    res = res.astype(int)
                if out.dtype == bool:
                    out = out.astype(int)
                out[evt_ids_ch[limarr]] += res[limarr]
            else:
                if res.dtype != bool:
                    res = res.astype(bool)

                if mode == "any":
                    out[evt_ids_ch] |= res & limarr

                if mode == "all":
                    out[evt_ids_ch] &= res & limarr

    return types.Array(nda=out)


def evaluate_at_channel(
    datainfo,
    tcm,
    channels,
    channels_skip,
    expr,
    field_list,
    ch_comp,
    pars_dict=None,
    default_value=np.nan,
    channel_mapping=None,
) -> types.Array:
    """Aggregates by evaluating the expression at a given channel.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    channels
        list of channels to be included for evaluation.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    ch_comp
       array of rawids at which the expression is evaluated.
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)
    table_id_fmt = datainfo.hit.table_fmt

    out = None

    for table_id in np.unique(ch_comp.nda.astype(int)):
        table_name = utils.get_table_name_by_pattern(table_id_fmt, table_id)
        # skip default value
        if table_name not in lh5.ls(datainfo.hit.file):
            continue

        # get index list for this channel to be loaded
        chan_tcm_indexs = ak.flatten(tcm.table_key) == table_id
        idx_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()

        evt_ids_ch = np.repeat(
            np.arange(0, len(tcm.table_key)), ak.sum(tcm.table_key == table_id, axis=1)
        )

        if (table_name in channels) and (table_name not in channels_skip):
            res = utils.get_data_at_channel(
                datainfo=datainfo,
                ch=table_name,
                tcm=tcm,
                expr=expr,
                field_list=field_list,
                pars_dict=pars_dict,
            )
        else:
            res = np.full(len(idx_ch), default_value)

        if out is None:
            out = utils.make_numpy_full(len(ch_comp.nda), default_value, res.dtype)

        out[evt_ids_ch] = np.where(
            table_id == ch_comp.nda[idx_ch], res, out[evt_ids_ch]
        )

    return types.Array(nda=out)


def evaluate_at_channel_vov(
    datainfo,
    tcm,
    expr,
    field_list,
    ch_comp,
    channels,
    channels_skip,
    pars_dict=None,
    default_value=np.nan,
    channel_mapping=None,
) -> types.VectorOfVectors:
    """Same as :func:`evaluate_at_channel` but evaluates expression at non
    flat channels :class:`.VectorOfVectors`.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    ch_comp
       array of "rawid"s at which the expression is evaluated.
    channels
       list of channels to be included for evaluation.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    ch_comp_channels = np.unique(ch_comp.flattened_data.nda).astype(int)

    out = np.full(
        len(ch_comp.flattened_data.nda), default_value, dtype=type(default_value)
    )

    type_name = None
    for table_id in ch_comp_channels:
        table_name = utils.get_table_name_by_pattern(datainfo.hit.table_fmt, table_id)
        evt_ids_ch = np.repeat(
            np.arange(0, len(tcm.table_key)), ak.sum(tcm.table_key == table_id, axis=1)
        )
        if (table_name in channels) and (table_name not in channels_skip):
            res = utils.get_data_at_channel(
                datainfo=datainfo,
                ch=table_name,
                tcm=tcm,
                expr=expr,
                field_list=field_list,
                pars_dict=pars_dict,
            )
            new_evt_ids_ch = np.repeat(
                np.arange(0, len(ch_comp)),
                ak.sum(ch_comp.view_as("ak") == table_id, axis=1),
            )
            matches = np.isin(evt_ids_ch, new_evt_ids_ch)
            out[ch_comp.flattened_data.nda == table_id] = res[matches]

        else:
            length = len(np.where(ch_comp.flattened_data.nda == table_id)[0])
            res = np.full(length, default_value)
            out[ch_comp.flattened_data.nda == table_id] = res

        if table_id == ch_comp_channels[0]:
            out = out.astype(res.dtype)
            type_name = res.dtype

    return types.VectorOfVectors(
        flattened_data=types.Array(out, dtype=type_name),
        cumulative_length=ch_comp.cumulative_length,
    )


def evaluate_to_aoesa(
    datainfo,
    tcm,
    channels,
    channels_skip,
    expr,
    field_list,
    query,
    n_rows,
    pars_dict=None,
    default_value=np.nan,
    missing_value=np.nan,
    channel_mapping=None,
) -> types.ArrayOfEqualSizedArrays:
    """Aggregates by returning an :class:`.ArrayOfEqualSizedArrays` of evaluated
    expressions of channels that fulfill a query expression.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    channels
       list of channels to be aggregated.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    query
       query expression to mask aggregation.
    n_rows
       length of output :class:`.VectorOfVectors`.
    ch_comp
       array of "rawid"s at which the expression is evaluated.
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    missing_value
       missing value.
    sorter
       sorts the entries in the vector according to sorter expression.
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    # define dimension of output array
    dtype = None
    out = None

    for i, ch in enumerate(channels):
        table_id = utils.get_tcm_id_by_pattern(datainfo.hit.table_fmt, ch)
        if table_id is None:
            continue

        # get index list for this channel to be loaded
        chan_tcm_indexs = ak.flatten(tcm.table_key) == table_id
        idx_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()

        evt_ids_ch = np.repeat(
            np.arange(0, len(tcm.table_key)), ak.sum(tcm.table_key == table_id, axis=1)
        )

        if ch not in channels_skip:
            res = utils.get_data_at_channel(
                datainfo=datainfo,
                ch=ch,
                tcm=tcm,
                expr=expr,
                field_list=field_list,
                pars_dict=pars_dict,
            )

            if dtype is None:
                dtype = res.dtype

            if out is None:
                out = utils.make_numpy_full(
                    (n_rows, len(channels)), missing_value, res.dtype
                )
        else:
            res = np.full(len(idx_ch), default_value)

        # get mask from query
        limarr = utils.get_mask_from_query(
            datainfo=datainfo,
            query=query,
            length=len(res),
            ch=ch,
            idx_ch=idx_ch,
        )

        out[evt_ids_ch, i] = np.where(limarr, res, out[evt_ids_ch, i])

    return out, dtype


def evaluate_to_vector(
    datainfo,
    tcm,
    channels,
    channels_skip,
    expr,
    field_list,
    query,
    n_rows,
    pars_dict=None,
    default_value=np.nan,
    sorter=None,
    channel_mapping=None,
) -> types.VectorOfVectors:
    """Aggregates by returning a :class:`.VectorOfVector` of evaluated
    expressions of channels that fulfill a query expression.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found.
    tcm
        TCM data arrays in an object that can be accessed by attribute.
    channels
       list of channels to be aggregated.
    channels_skip
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    field_list
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    query
       query expression to mask aggregation.
    n_rows
       length of output :class:`.VectorOfVectors`.
    ch_comp
       array of "rawids" at which the expression is evaluated.
    pars_dict
       dictionary of `evt` and additional parameters and their values.
    default_value
       default value.
    sorter
       sorts the entries in the vector according to sorter expression.
       ``ascend_by:<hit|dsp.field>`` results in an vector ordered ascending,
       ``decend_by:<hit|dsp.field>`` sorts descending.
    """
    out, dtype = evaluate_to_aoesa(
        datainfo=datainfo,
        tcm=tcm,
        channels=channels,
        channels_skip=channels_skip,
        expr=expr,
        field_list=field_list,
        query=query,
        n_rows=n_rows,
        pars_dict=pars_dict,
        default_value=default_value,
        missing_value=np.nan,
    )

    # if a sorter is given sort accordingly
    if sorter is not None:
        md, fld = sorter.split(":")
        s_val, _ = evaluate_to_aoesa(
            datainfo=datainfo,
            tcm=tcm,
            channels=channels,
            channels_skip=channels_skip,
            expr=fld,
            field_list=[tuple(fld.split("."))],
            query=None,
            n_rows=n_rows,
            missing_value=np.nan,
        )

        if "ascend_by" == md:
            out = out[np.arange(len(out))[:, None], np.argsort(s_val)]

        elif "descend_by" == md:
            out = out[np.arange(len(out))[:, None], np.argsort(-s_val)]
        else:
            raise ValueError(
                "sorter values can only have 'ascend_by' or 'descend_by' prefixes"
            )

    return types.VectorOfVectors(
        ak.values_astype(ak.drop_none(ak.nan_to_none(ak.Array(out))), dtype)
    )



"""
Subclasses of scipy.stat's rv_continuous, allows for freezing of pygama numba-fied distributions for faster functions
Example:

>>> moyal = moyal_gen(name='moyal')
>>> moyal = moyal(1, 2)
>>> moyal.get_pdf([-1,2,3]) # a direct call to the faster numbafied method
>>> moyal.pdf([-1,2,3]) # a call to the slower scipy method, allows access to other scipy methods like .rvs
>>> moyal.rvs(100) # Can access scipy methods!

NOTE: the dist.pdf method is a slow scipy method, and the dist.get_pdf method is fast
"""

import numpy as np
from scipy.stats._distn_infrastructure import rv_continuous, rv_frozen


class NumbaFrozen(rv_frozen):
    r"""
    Essentially, an overloading of the definition of rv_continuous_frozen so that our pygama class instantiations have
    access to both the slower scipy methods, as well as the faster pygama methods (like get_pdf)
    """

    def pdf(self, x: np.ndarray) -> np.ndarray:
        r"""
        Overloading of scipy's pdf function, this is a slow function
        """

        return self.dist.pdf(x, *self.args, **self.kwds)

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        r"""
        Overloading of scipy's logpdf function, this is a slow function
        """

        return self.dist.logpdf(x, *self.args, **self.kwds)

    def get_pdf(self, x: np.ndarray) -> np.ndarray:
        r"""
        Direct access to numba-fied pdfs, a fast function

        Returns
        -------
        support normalized pdf
        """

        return self.dist.get_pdf(np.array(x), *self.args, **self.kwds)

    def get_cdf(self, x: np.ndarray) -> np.ndarray:
        r"""
        Direct access to numba-fied cdfs, a fast function

        Returns
        -------
        support normalized cdf
        """

        return self.dist.get_cdf(np.array(x), *self.args, **self.kwds)

    def pdf_ext(
        self, x: np.ndarray, x_lo: float, x_hi: float, area: float
    ) -> tuple[float, np.ndarray]:
        r"""
        Direct access to numba-fied extended pdfs, a fast function

        Returns
        -------
        integral, scaled support normalized pdf
        """

        return self.dist.pdf_ext(np.array(x), x_lo, x_hi, area, *self.args, **self.kwds)

    def cdf_ext(self, x: np.ndarray, area: float) -> np.ndarray:
        r"""
        Direct access to numba-fied extended pdfs, a fast function

        Returns
        -------
        scaled support normalized cdf
        """

        return self.dist.cdf_ext(np.array(x), area, *self.args, **self.kwds)

    def pdf_norm(self, x, x_lo: float, x_hi: float):
        r"""
        Direct access to numba-fied pdfs, a fast function. Normalized on a fit range

        Returns
        -------
        fit-range normalized pdf
        """
        return self.dist.pdf_norm(x, x_lo, x_hi, *self.args, **self.kwds)

    def cdf_norm(self, x, x_lo: float, x_hi: float):
        r"""
        Direct access to numba-fied cdfs, a fast function. Normalized on a fit range

        Returns
        -------
        fit-range normalized cdf
        """
        return self.dist.cdf_norm(x, x_lo, x_hi, *self.args, **self.kwds)

    def required_args(self) -> tuple:
        r"""
        Allow access to the required args of frozen distribution

        Returns
        -------
        Required shape, location, and scale arguments
        """

        return self.dist.required_args()


class PygamaContinuous(rv_continuous):
    r"""
    Subclass rv_continuous, and modify the instantiation so that we call an overloaded
    version of rv_continuous_frozen that has direct access to pygama numbafied functions
    """

    def _pdf_norm(self, x, x_lo, x_hi, *args, **kwds):
        r"""
        Normalize a pdf on a subset of its support, typically over a fit-range.

        Parameters
        ----------
        x
            The data
        x_lo
            The lower range to normalize on
        x_hi
            The upper range to normalize on
        args
            The shape and location and scale parameters of a specific distribution

        Returns
        -------
        pdf_norm
            The pdf that is normalized on a smaller range

        Notes
        -----
        If upper_range and lower_range are both :func:`np.inf`, then the function automatically takes x_lo =:func:`np.amin(x)`, x_hi=:func:`np.amax(x)`
        We also need to overload this in every subclass because we want functions to be able to introspect the shape and location/scale names
        For distributions that are only defined on a limited range, like with lower_range, upper_range, we don't need to call these, instead just call the normal pdf.
        """

        norm = np.diff(self.get_cdf(np.array([x_lo, x_hi]), *args))

        if norm == 0:
            return np.full_like(x, np.inf)
        else:
            return self.get_pdf(x, *args) / norm

    def _cdf_norm(self, x, x_lo, x_hi, *args, **kwds):
        r"""
        Derive a cdf from a pdf that is normalized on a subset of its support, typically over a fit-range.

        Parameters
        ----------
        x
            The data
        x_lo
            The lower range to normalize on
        x_hi
            The upper range to normalize on
        args
            The shape and location and scale parameters of a specific distribution

        Returns
        -------
        pdf_norm
            The pdf that is normalized on a smaller range

        Notes
        -----
        If upper_range and lower_range are both :func:`np.inf`, then the function automatically takes x_lo =:func:`np.amin(x)`, x_hi=:func:`np.amax(x)`
        We also need to overload this in every subclass because we want functions to be able to introspect the shape and location/scale names.
        For distributions that are only defined on a limited range, like with lower_range, upper_range, we don't need to call these, instead just call the normal pdf.
        """

        norm = np.diff(self.get_cdf(np.array([x_lo, x_hi]), *args))

        if norm == 0:
            return np.full_like(x, np.inf)
        else:
            return (self.get_cdf(x, *args)) / norm

    def __call__(self, *args, **kwds):
        return NumbaFrozen(self, *args, **kwds)



r"""
Provide a convenience function for the HPGe peak shape.

A HPGe peak consists of a Gaussian
on an Exgauss on a step function.

.. math::

    PDF = n_sig*((1-htail)*gauss + htail*exgauss) + n_bkg*step


Called with

hpge_peak.get_pdf(x, x_lo, x_hi, n_sig, mu, sigma, htail, tau, n_bkg, hstep)

Parameters
----------
x_lo
    Lower bound of the step function
x_hi
    Upper bound of the step function
n_sig
    The area of the gauss on exgauss
mu
    The centroid of the Gaussian
sigma
    The standard deviation of the Gaussian
htail
    The height of the Gaussian tail
tau
    The characteristic scale of the Gaussian tail
n_bkg
    The area of the step background
hstep
    The height of the step function background

Returns
-------
hpge_peak
    A subclass of SumDists and rv_continuous, has methods of pdf, cdf, etc.

Notes
-----
The extended Gaussian distribution and the step distribution share the mu, sigma with the Gaussian
"""

import numpy as np

from pygama.math.functions.gauss_on_exgauss import gauss_on_exgauss
from pygama.math.functions.step import step
from pygama.math.functions.sum_dists import SumDists
from pygama.math.hpge_peak_fitting import hpge_peak_fwfm, hpge_peak_fwhm, hpge_peak_mode

(x_lo, x_hi, n_sig, mu, sigma, frac1, tau, n_bkg, hstep) = range(9)
par_array = [
    (gauss_on_exgauss, [mu, sigma, frac1, tau]),
    (step, [x_lo, x_hi, mu, sigma, hstep]),
]

hpge_peak = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=[
        "x_lo",
        "x_hi",
        "n_sig",
        "mu",
        "sigma",
        "htail",
        "tau",
        "n_bkg",
        "hstep",
    ],
    name="hpge_peak",
)


# This is defined here as to avoid a circular import inside `SumDists`
def hpge_get_fwhm(self, pars: np.ndarray, cov: np.ndarray = None) -> tuple:
    r"""
    Get the fwhm value from the output of a fit quickly
    Need to overload this to use hpge_peak_fwhm (to avoid a circular import) for when self is an hpge peak,
    and otherwise returns 2sqrt(2log(2))*sigma

    Parameters
    ----------
    pars
        Array of fit parameters
    cov
        Optional, array of covariances for calculating error on the fwhm


    Returns
    -------
    fwhm, error
        the value of the fwhm and its error
    """
    req_args = np.array(self.required_args())
    sigma_idx = np.where(req_args == "sigma")[0][0]

    if ("htail" in req_args) and (
        "hstep" in req_args
    ):  # having both the htail and hstep means it is an exgauss on a step
        htail_idx = np.where(req_args == "htail")[0][0]
        tau_idx = np.where(req_args == "tau")[0][0]
        # We need to ditch the x_lo and x_hi columns and rows
        if cov is not None:
            cov = np.array(cov)
            dropped_cov = cov[:, 2:][2:, :]
            return hpge_peak_fwhm(
                pars[sigma_idx], pars[htail_idx], pars[tau_idx], dropped_cov
            )
        else:
            return hpge_peak_fwhm(pars[sigma_idx], pars[htail_idx], pars[tau_idx])

    else:
        if cov is None:
            return pars[sigma_idx] * 2 * np.sqrt(2 * np.log(2))
        else:
            return pars[sigma_idx] * 2 * np.sqrt(2 * np.log(2)), np.sqrt(
                cov[sigma_idx][sigma_idx]
            ) * 2 * np.sqrt(2 * np.log(2))


# This is defined here as to avoid a circular import inside `SumDists`
def hpge_get_fwfm(
    self, pars: np.ndarray, frac_max=0.5, cov: np.ndarray = None
) -> tuple:
    r"""
    Get the fwhm value from the output of a fit quickly
    Need to overload this to use hpge_peak_fwhm (to avoid a circular import) for when self is an hpge peak,
    and otherwise returns 2sqrt(-2log(frac_max))*sigma

    Parameters
    ----------
    pars
        Array of fit parameters
    cov
        Optional, array of covariances for calculating error on the fwhm


    Returns
    -------
    fwhm, error
        the value of the fwhm and its error
    """
    req_args = np.array(self.required_args())
    sigma_idx = np.where(req_args == "sigma")[0][0]

    if ("htail" in req_args) and (
        "hstep" in req_args
    ):  # having both the htail and hstep means it is an exgauss on a step
        htail_idx = np.where(req_args == "htail")[0][0]
        tau_idx = np.where(req_args == "tau")[0][0]
        # We need to ditch the x_lo and x_hi columns and rows
        if cov is not None:
            cov = np.array(cov)
            dropped_cov = cov[:, 2:][2:, :]
            return hpge_peak_fwfm(
                pars[sigma_idx],
                pars[htail_idx],
                pars[tau_idx],
                frac_max=frac_max,
                cov=dropped_cov,
            )
        else:
            return hpge_peak_fwfm(
                pars[sigma_idx], pars[htail_idx], pars[tau_idx], frac_max=frac_max
            )

    else:
        if cov is None:
            return pars[sigma_idx] * 2 * np.sqrt(2 * np.log(2))
        else:
            return pars[sigma_idx] * 2 * np.sqrt(-2 * np.log(frac_max)), np.sqrt(
                cov[sigma_idx][sigma_idx]
            ) * 2 * np.sqrt(-2 * np.log(frac_max))


# This is defined here as to avoid a circular import inside `SumDists`
def hpge_get_mode(self, pars: np.ndarray, cov: np.ndarray = None) -> tuple:
    r"""
    Get the fwhm value from the output of a fit quickly
    Need to overload this to use hpge_peak_fwhm (to avoid a circular import) for when self is an hpge peak,
    and otherwise returns 2sqrt(2log(2))*sigma

    Parameters
    ----------
    pars
        Array of fit parameters
    cov
        Optional, array of covariances for calculating error on the fwhm


    Returns
    -------
    fwhm, error
        the value of the fwhm and its error
    """
    req_args = np.array(self.required_args())
    sigma_idx = np.where(req_args == "sigma")[0][0]
    mu_idx = np.where(req_args == "mu")[0][0]

    if ("htail" in req_args) and (
        "hstep" in req_args
    ):  # having both the htail and hstep means it is an exgauss on a step
        htail_idx = np.where(req_args == "htail")[0][0]
        tau_idx = np.where(req_args == "tau")[0][0]
        # We need to ditch the x_lo and x_hi columns and rows
        if cov is not None:
            cov = np.array(cov)
            dropped_cov = cov[2:, 2:]

            return hpge_peak_mode(
                pars[mu_idx],
                pars[sigma_idx],
                pars[htail_idx],
                pars[tau_idx],
                dropped_cov,
            )
        else:
            return hpge_peak_mode(
                pars[mu_idx], pars[sigma_idx], pars[htail_idx], pars[tau_idx]
            )

    else:
        if cov is None:
            return pars[mu_idx]
        else:
            return pars[mu_idx], np.sqrt(cov[mu_idx][mu_idx])


# hpge_peak.get_fwhm = hpge_get_fwhm
hpge_peak.get_fwfm = hpge_get_fwfm.__get__(hpge_peak)
hpge_peak.get_mode = hpge_get_mode.__get__(hpge_peak)
hpge_peak.get_fwhm = hpge_get_fwhm.__get__(hpge_peak)



import numpy as np
from scipy.stats import uniform as scipy_uniform

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.math.functions.uniform import uniform


def test_uniform_pdf():
    x = np.arange(-10, 12)
    a = -1
    b = 10

    y = uniform.pdf(x, a, b)
    y_direct = uniform.get_pdf(x, a, b)
    scipy_y = scipy_uniform.pdf(x, a, b)

    assert isinstance(uniform, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.array_equal(y_direct, scipy_y)

    n_sig = 20
    x_lo = -1
    x_hi = 9

    y_sig, y_ext = uniform.pdf_ext(x, x_lo, x_hi, n_sig, a, b)
    assert np.array_equal(y_ext, n_sig * scipy_y)
    assert np.array_equal(y_sig, n_sig)

    y_norm = uniform.pdf_norm(x, x[0], x[-1], a, b)

    assert np.allclose(y_norm, y_direct, rtol=1e-8)


def test_uniform_cdf():
    x = np.arange(-10, 12)
    a = -1
    b = 10

    y = uniform.cdf(x, a, b)
    y_direct = uniform.get_cdf(x, a, b)
    scipy_y = scipy_uniform.cdf(x, a, b)

    assert isinstance(uniform, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = uniform.cdf_ext(x, n_sig, a, b)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    y_norm = uniform.cdf_norm(x, x[0], x[-1], a, b)

    assert np.allclose(y_norm, y_direct, rtol=1e-8)


def test_required_args():
    names = uniform.required_args()
    assert names[0] == "a"
    assert names[1] == "b"


def test_name():
    assert uniform.name == "uniform"



"""
This module provides routines for calculating and applying quality cuts
"""

from __future__ import annotations

import logging
import re

import awkward as ak
import lgdo.lh5 as lh5
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from lgdo.types import Table
from scipy import stats
from scipy.stats import chi2, skewnorm

import pygama.math.binned_fitting as pgf
import pygama.math.histogram as pgh
import pygama.pargen.energy_cal as pgc
from pygama.math.binned_fitting import goodness_of_fit
from pygama.math.distributions import exgauss, gaussian
from pygama.math.functions.sum_dists import SumDists
from pygama.math.unbinned_fitting import fit_unbinned

(x_lo, x_hi, n_sig, mu, sigma, n_bkg, tau) = range(7)
par_array = [(gaussian, [mu, sigma]), (exgauss, [mu, sigma, tau])]
gauss_on_exgauss_areas = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "n_bkg", "tau"],
    name="gauss_on_exgauss_areas",
)

(x_lo, x_hi, n_sig, mu, sigma, tau1, n_bkg, tau2) = range(8)
par_array = [(exgauss, [mu, sigma, tau1]), (exgauss, [mu, sigma, tau2])]
double_exgauss = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "tau1", "n_bkg", "tau2"],
    name="double_exgauss",
)


def skewed_fit(x, n_sig, mu, sigma, alpha):
    return n_sig, n_sig * skewnorm.pdf(x, alpha, mu, sigma)


def skewed_pdf(x, n_sig, mu, sigma, alpha):
    return n_sig * skewnorm.pdf(x, alpha, mu, sigma)


log = logging.getLogger(__name__)
sto = lh5.LH5Store()
mpl.use("agg")


def get_keys(in_data, cut_dict):
    """
    Get the keys of the data that are used in the cut dictionary
    """
    parameters = []
    for _, entry in cut_dict.items():
        if "cut_parameter" in entry:
            parameters.append(entry["cut_parameter"])
        else:
            parameters.append(entry["expression"])

    out_params = []
    if isinstance(in_data, dict):
        possible_keys = in_data.keys()
    elif isinstance(in_data, list):
        possible_keys = in_data
    for param in parameters:
        for key in possible_keys:
            if key in param:
                out_params.append(key)
    return np.unique(out_params).tolist()


def get_mode_stdev(par_array):
    idxs = (par_array > np.nanpercentile(par_array, 1)) & (
        par_array < np.nanpercentile(par_array, 99)
    )
    par_array = par_array[idxs]
    bin_width = np.nanpercentile(par_array, 55) - np.nanpercentile(par_array, 50)

    counts, start_bins, var = pgh.get_hist(
        par_array,
        range=(np.nanmin(par_array), np.nanmax(par_array)),
        dx=bin_width,
    )
    max_idx = np.argmax(counts)
    mu = start_bins[max_idx]
    try:
        fwhm = pgh.get_fwhm(counts, start_bins)[0]
        guess_sig = fwhm / 2.355

        lower_bound = mu - 10 * guess_sig

        upper_bound = mu + 10 * guess_sig

    except Exception:
        lower_bound = np.nanpercentile(par_array, 5)
        upper_bound = np.nanpercentile(par_array, 95)

    if (lower_bound < np.nanmin(par_array)) or (lower_bound > np.nanmax(par_array)):
        lower_bound = np.nanmin(par_array)
    if (upper_bound > np.nanmax(par_array)) or (upper_bound < np.nanmin(par_array)):
        upper_bound = np.nanmax(par_array)

    try:
        counts, bins, var = pgh.get_hist(
            par_array,
            dx=(np.nanpercentile(par_array, 52) - np.nanpercentile(par_array, 50)),
            range=(lower_bound, upper_bound),
        )

        bin_centres = pgh.get_bin_centers(bins)

        fwhm = pgh.get_fwhm(counts, bins)[0]
        mean = float(bin_centres[np.argmax(counts)])
        pars, cov = pgf.gauss_mode_width_max(
            counts,
            bins,
            mode_guess=mean,
            n_bins=20,
            cost_func="Least Squares",
            inflate_errors=False,
            gof_method="var",
        )
        mean = pars[0]
        std = fwhm / 2.355

        if (
            mean < np.nanmin(bins)
            or mean > np.nanmax(bins)
            or (mean + std) < mu
            or (mean - std) > mu
        ):
            raise IndexError
    except IndexError:
        try:
            fwhm = pgh.get_fwhm(counts, bins)[0]
            mean = float(bin_centres[np.argmax(counts)])
            std = fwhm / 2.355
        except Exception:
            lower_bound = np.nanpercentile(par_array, 5)
            upper_bound = np.nanpercentile(par_array, 95)

            counts, bins, var = pgh.get_hist(
                par_array,
                dx=np.nanpercentile(par_array, 52) - np.nanpercentile(par_array, 50),
                range=(lower_bound, upper_bound),
            )

            bin_centres = pgh.get_bin_centers(bins)

            try:
                fwhm = pgh.get_fwhm(counts, bins)[0]
                mean = float(bin_centres[np.argmax(counts)])
                std = fwhm / 2.355
            except Exception:
                mean = float(bin_centres[np.argmax(counts)])
                std = np.nanstd(par_array)
    return mean, std


def fit_distributions(x_lo, x_hi, norm_par_array, display=0):
    peak_par_array = norm_par_array[(norm_par_array > x_lo) & (norm_par_array < x_hi)]

    hist, bins, var = pgh.get_hist(peak_par_array, dx=0.1, range=(x_lo, x_hi))
    var = np.where(var == 0, 1, var)

    exgauss_pars, _, _ = fit_unbinned(
        exgauss.pdf_ext,
        peak_par_array,
        [x_lo, x_hi, len(peak_par_array), 0, 1, -0.1],
        simplex=True,
        bounds=[
            (None, None),
            (None, None),
            (0, None),
            (-0.8, 0.8),
            (0.8, 1.2),
            (None, None),
        ],
        fixed=["x_lo", "x_hi"],
    )

    gauss_pars, _, _ = fit_unbinned(
        gaussian.pdf_ext,
        peak_par_array,
        [x_lo, x_hi, len(peak_par_array), 0, 1],
        simplex=True,
        bounds=[(None, None), (None, None), (0, None), (-0.5, 0.5), (0.5, 1.5)],
        fixed=["x_lo", "x_hi"],
    )

    gauss_on_exgauss_pars, _, _ = fit_unbinned(
        gauss_on_exgauss_areas.pdf_ext,
        peak_par_array,
        [x_lo, x_hi, len(peak_par_array) * 0.9, 0, 1, len(peak_par_array) * 0.1, -0.1],
        simplex=True,
        bounds=[
            (None, None),
            (None, None),
            (0, None),
            (-0.5, 0.5),
            (0, None),
            (0, None),
            (None, None),
        ],
        fixed=["x_lo", "x_hi"],
    )

    skewed_pars, _, _ = fit_unbinned(
        skewed_fit,
        peak_par_array,
        [len(peak_par_array), 0, 1, 0.1],
        simplex=True,
        bounds=[(0, None), (None, None), (0, None), (None, None)],
    )

    double_exgauss_pars, _, _ = fit_unbinned(
        double_exgauss.pdf_ext,
        peak_par_array,
        [
            x_lo,
            x_hi,
            len(peak_par_array) * 0.5,
            0,
            1,
            -0.1,
            len(peak_par_array) * 0.5,
            0.1,
        ],
        simplex=True,
        bounds=[
            (None, None),
            (None, None),
            (0, None),
            (-0.5, 0.5),
            (0, None),
            (None, 0),
            (0, None),
            (0, None),
        ],
        fixed=["x_lo", "x_hi"],
    )

    gauss_csqr = goodness_of_fit(
        hist,
        bins,
        var,
        lambda x, *args: gaussian.pdf_ext(x, *args)[1],
        gauss_pars,
        method="var",
        scale_bins=True,
    )

    exgauss_csqr = goodness_of_fit(
        hist,
        bins,
        var,
        lambda x, *args: exgauss.pdf_ext(x, *args)[1],
        exgauss_pars,
        method="var",
        scale_bins=True,
    )

    skewed_csqr = goodness_of_fit(
        hist,
        bins,
        var,
        lambda x, *args: skewed_fit(x, *args)[1],
        skewed_pars,
        method="var",
        scale_bins=True,
    )

    gauss_on_exgauss_csqr = goodness_of_fit(
        hist,
        bins,
        var,
        gauss_on_exgauss_areas.get_pdf,
        gauss_on_exgauss_pars,
        method="var",
        scale_bins=True,
    )

    double_exgauss_csqr = goodness_of_fit(
        hist,
        bins,
        var,
        double_exgauss.get_pdf,
        double_exgauss_pars,
        method="var",
        scale_bins=True,
    )

    if display > 0:
        bcs = pgh.get_bin_centers(bins)
        plt.figure()
        plt.step(bcs, hist)
        plt.plot(
            bcs, double_exgauss.get_pdf(bcs, *double_exgauss_pars) * np.diff(bins)[0]
        )
        plt.plot(
            bcs,
            gauss_on_exgauss_areas.pdf_ext(bcs, *gauss_on_exgauss_pars)[1]
            * np.diff(bins)[0],
        )
        plt.plot(bcs, skewed_fit(bcs, *skewed_pars)[1] * np.diff(bins)[0])
        plt.plot(bcs, gaussian.pdf_ext(bcs, *gauss_pars)[1] * np.diff(bins)[0])
        plt.plot(bcs, exgauss.pdf_ext(bcs, *exgauss_pars)[1] * np.diff(bins)[0])
        plt.show()

    gauss_p_val = chi2.sf(gauss_csqr[0], gauss_csqr[1] + 2)
    exgauss_p_val = chi2.sf(exgauss_csqr[0], exgauss_csqr[1] + 2)
    skewed_p_val = chi2.sf(skewed_csqr[0], skewed_csqr[1])
    gauss_on_exgauss_p_val = chi2.sf(
        gauss_on_exgauss_csqr[0], gauss_on_exgauss_csqr[1] + 2
    )
    double_exgauss_p_val = chi2.sf(double_exgauss_csqr[0], double_exgauss_csqr[1] + 2)

    funcs = [gaussian, exgauss, skewed_fit, gauss_on_exgauss_areas, double_exgauss]
    pars = [
        gauss_pars,
        exgauss_pars,
        skewed_pars,
        gauss_on_exgauss_pars,
        double_exgauss_pars,
    ]
    pvals = np.array(
        [
            gauss_p_val,
            exgauss_p_val,
            skewed_p_val,
            gauss_on_exgauss_p_val,
            double_exgauss_p_val,
        ]
    )
    csqrs = [
        gauss_csqr[0],
        exgauss_csqr[0],
        skewed_csqr[0],
        gauss_on_exgauss_csqr[0],
        double_exgauss_csqr[0],
    ]

    if (pvals == 0).all():
        idx = np.nanargmin(csqrs)
    else:
        idx = np.nanargmax(pvals)
    func = funcs[idx]
    pars = pars[idx]
    return func, pars


def generate_cuts(
    data: dict[str, np.ndarray],
    cut_dict: dict[str, int],
    rounding: int = 4,
    display: int = 0,
) -> dict:
    """
    Finds double sided cut boundaries for a file for the parameters specified

    Parameters
    ----------
    data : lh5 table, dictionary of arrays or pandas dataframe
        data to calculate cuts on

    parameters : dict
        dictionary of the form:

        .. code-block:: json

            {
                "output_parameter_name": {
                    "cut_parameter": "parameter_to_cut_on",
                    "cut_level": "number_of_sigmas",
                    "mode": "inclusive/exclusive"
                }
            }

        number of sigmas can instead be a dictionary to specify different cut levels for low and high side
        or to only have a one sided cut only specify one of the low or high side
        e.g.

        .. code-block:: json

            {
                "output_parameter_name": {
                    "cut_parameter": "parameter_to_cut_on",
                    "cut_level": {"low_side": "3", "high_side": "2"},
                    "mode": "inclusive/exclusive"
                }
            }

        alternatively can specify hit dict fields to just copy dict into output dict e.g.

        .. code-block:: json

            {
                "is_valid_t0":{
                    "expression":"(tp_0_est>a)&(tp_0_est<b)",
                    "parameters":{"a":"46000", "b":"52000"}
                }
            }

        or

        .. code-block:: json

            {
                "is_valid_cal":{
                    "expression":"(~is_pileup_tail)&(~is_pileup_baseline)"
                }
            }

    rounding : int
        number of decimal places to round to

    display : int
        if 1 will display plots of the cuts
        if 0 will not display plots

    Returns
    -------
    dict
        dictionary of the form (same as hit dicts):

        .. code-block:: python

            {
                "output_parameter_name": {
                    "expression": "cut_expression",
                    "parameters": {"a": "lower_bound", "b": "upper_bound"}
                }
            }

    plot_dict
        dictionary of plots
    """

    output_dict = {}
    plot_dict = {}
    if isinstance(data, pd.DataFrame):
        pass
    elif isinstance(data, Table):
        data = {entry: data[entry].nda for entry in get_keys(data, cut_dict)}
        data = pd.DataFrame.from_dict(data)
    elif isinstance(data, dict):
        data = pd.DataFrame.from_dict(data)
    for out_par, cut in cut_dict.items():
        if "expression" in cut:
            output_dict[out_par] = {"expression": cut["expression"]}
            if "parameters" in cut:
                output_dict[out_par].update({"parameters": cut["parameters"]})
        else:
            par = cut["cut_parameter"]
            num_sigmas = cut["cut_level"]
            mode = cut["mode"]
            try:
                all_par_array = data[par].to_numpy()
            except KeyError:
                all_par_array = data.eval(par).to_numpy()

            mean, std = get_mode_stdev(all_par_array)

            if isinstance(num_sigmas, (int, float)):
                num_sigmas_left = num_sigmas
                num_sigmas_right = num_sigmas
            elif isinstance(num_sigmas, dict):
                if "low_side" in num_sigmas:
                    num_sigmas_left = num_sigmas["low_side"]
                else:
                    num_sigmas_left = None
                if "high_side" in num_sigmas:
                    num_sigmas_right = num_sigmas["high_side"]
                else:
                    num_sigmas_right = None
            upper = round(float((num_sigmas_right * std) + mean), rounding)
            lower = round(float((-num_sigmas_left * std) + mean), rounding)
            if mode == "inclusive":
                if upper is not None and lower is not None:
                    cut_string = f"({par}>a) & ({par}<b)"
                    par_dict = {"a": lower, "b": upper}
                elif upper is None:
                    cut_string = f"{par}>a"
                    par_dict = {"a": lower}
                elif lower is None:
                    cut_string = f"{par}<a"
                    par_dict = {"a": upper}
            elif mode == "exclusive":
                if upper is not None and lower is not None:
                    cut_string = f"({par}<a) | ({par}>b)"
                    par_dict = {"a": lower, "b": upper}
                elif upper is None:
                    cut_string = f"{par}<a"
                    par_dict = {"a": lower}
                elif lower is None:
                    cut_string = f"{par}>a"
                    par_dict = {"a": upper}

            output_dict[out_par] = {"expression": cut_string, "parameters": par_dict}

            if display > 0:
                fig = plt.figure()
                low_val = np.nanpercentile(all_par_array, 5)
                up_val = np.nanpercentile(all_par_array, 95)
                if upper is not None:
                    plt.axvline(upper)
                    if up_val < upper:
                        up_val = upper
                if lower is not None:
                    plt.axvline(lower)
                    if low_val > lower:
                        low_val = lower

                plt.hist(
                    all_par_array,
                    bins=np.linspace(
                        low_val,
                        up_val,
                        100,
                    ),
                    histtype="step",
                )

                plt.ylabel("counts")
                plt.xlabel(out_par)
                plot_dict[out_par] = fig
                plt.close()
    if display > 0:
        return output_dict, plot_dict
    else:
        return output_dict


def get_cut_indexes(data, cut_parameters):
    """
    Get the indexes of the data that pass the cuts in
    """
    cut_dict = generate_cuts(data, cut_dict=cut_parameters)
    log.debug(f"Cuts are {cut_dict}")

    if isinstance(data, Table):
        ct_mask = np.full(len(data), True, dtype=bool)
        for outname, info in cut_dict.items():
            outcol = data.eval(info["expression"], info.get("parameters", None))
            data.add_column(outname, outcol)
        log.debug("Applied Cuts")

        for cut in cut_dict:
            ct_mask = data[cut].nda & ct_mask
    elif isinstance(data, pd.DataFrame):
        ct_mask = np.full(len(data), True, dtype=bool)

        for outname, info in cut_dict.items():
            # convert to pandas eval
            exp = info["expression"]
            for key in info.get("parameters", None):
                exp = re.sub(f"(?<![a-zA-Z0-9]){key}(?![a-zA-Z0-9])", f"@{key}", exp)
            data[outname] = data.eval(exp, local_dict=info.get("parameters", None))
            ct_mask = ct_mask & data[outname]
    else:
        raise ValueError("Data must be a Table or DataFrame")

    return ct_mask


def generate_cut_classifiers(
    data: dict[str, np.ndarray],
    cut_dict: dict[str, int],
    rounding: int = 4,
    display: int = 0,
) -> dict:
    """
    Finds double sided cut boundaries for a file for the parameters specified

    Parameters
    ----------
    data : lh5 table, dictionary of arrays or pandas dataframe
        data to calculate cuts on

    parameters : dict
        dictionary of the form:

        .. code-block:: json

            {
                "output_parameter_name": {
                    "cut_parameter": "parameter_to_cut_on",
                    "cut_level": "number_of_sigmas",
                     "mode": "inclusive/exclusive"
                }
            }

        number of sigmas can instead be a dictionary to specify different cut levels for low and high side
        or to only have a one sided cut only specify one of the low or high side
        e.g.

        .. code-block:: json

            {
                "output_parameter_name": {
                    "cut_parameter": "parameter_to_cut_on",
                    "cut_level": {"low_side": "3", "high_side": "2"},
                    "mode": "inclusive/exclusive"
                }
            }

        alternatively can specify hit dict fields to just copy dict into output dict e.g.

        .. code-block:: json

            {
                "is_valid_t0":{
                    "expression":"(tp_0_est>a)&(tp_0_est<b)",
                    "parameters":{"a":"46000", "b":"52000"}
                }
            }

        or

        .. code-block:: json

            {
                "is_valid_cal":{
                    "expression":"(~is_pileup_tail)&(~is_pileup_baseline)"
                }
            }

    rounding : int
        number of decimal places to round to

    display : int
        if 1 will display plots of the cuts
        if 0 will not display plots

    Returns
    -------
    dict
        dictionary of the form (same as hit dicts):

        .. code-block:: python

            {
                "output_parameter_name": {
                    "expression": "cut_expression",
                    "parameters": {"a": lower_bound, "b": upper_bound}
                }
            }

    plot_dict
        dictionary of plots
    """

    output_dict = {}
    plot_dict = {}
    if isinstance(data, pd.DataFrame):
        pass
    elif isinstance(data, Table):
        data = {entry: data[entry].nda for entry in get_keys(data, cut_dict)}
        data = pd.DataFrame.from_dict(data)
    elif isinstance(data, dict):
        data = pd.DataFrame.from_dict(data)
    for out_par, cut in cut_dict.items():
        if "expression" in cut:
            output_dict[out_par] = {"expression": cut["expression"]}
            if "parameters" in cut:
                output_dict[out_par].update({"parameters": cut["parameters"]})
        else:
            par = cut["cut_parameter"]
            num_sigmas = cut.get("cut_level", None)
            percentile = cut.get("cut_percentile", None)
            default = cut.get("default", None)
            method = cut.get("method", "fit")
            mode = cut["mode"]
            try:
                all_par_array = data[par].to_numpy()
            except KeyError:
                all_par_array = data.eval(par).to_numpy()

            mean, std = get_mode_stdev(all_par_array)

            norm_par_array = (all_par_array - mean) / std

            if num_sigmas is not None:
                if isinstance(num_sigmas, (int, float)):
                    cut_left = -num_sigmas
                    cut_right = num_sigmas
                elif isinstance(num_sigmas, dict):
                    if "low_side" in num_sigmas:
                        cut_left = num_sigmas["low_side"]
                    else:
                        cut_left = None
                    if "high_side" in num_sigmas:
                        cut_right = num_sigmas["high_side"]
                    else:
                        cut_right = None

            elif percentile is not None:
                if method == "fit":
                    try:
                        x_lo = -10
                        x_hi = 10
                        func, pars = fit_distributions(x_lo, x_hi, norm_par_array)

                    except RuntimeError:
                        x_lo = -20
                        x_hi = 20
                        func, pars = fit_distributions(x_lo, x_hi, norm_par_array)

                    range_low, range_high = (-100, 100)
                    xs = np.arange(range_low, range_high, 0.1)
                    if func == exgauss:
                        cdf = exgauss.cdf_norm(
                            xs,
                            range_low,
                            range_high,
                            pars["mu"],
                            pars["sigma"],
                            pars["tau"],
                        )
                    elif func == gaussian:
                        cdf = gaussian.cdf_norm(
                            xs,
                            range_low,
                            range_high,
                            pars["mu"],
                            pars["sigma"],
                        )
                    elif func == gauss_on_exgauss_areas or func == double_exgauss:
                        cdf = func.cdf_norm(xs, range_low, range_high, *pars[2:])
                    elif func == skewed_fit:
                        cdf = skewnorm.cdf(xs, pars["alpha"], pars["mu"], pars["sigma"])
                    else:
                        raise ValueError("unknown func")

                    if isinstance(percentile, (int, float)):
                        cut_left = xs[np.argmin(np.abs(cdf - (1 - (percentile / 100))))]
                        cut_right = xs[np.argmin(np.abs(cdf - (percentile / 100)))]

                    elif isinstance(percentile, dict):
                        if "low_side" in percentile:
                            cut_left = xs[
                                np.argmin(np.abs(cdf - (1 - (percentile / 100))))
                            ]
                        else:
                            cut_left = None
                        if "high_side" in percentile:
                            cut_right = xs[np.argmin(np.abs(cdf - (percentile / 100)))]
                        else:
                            cut_right = None

                else:
                    if isinstance(percentile, (int, float)):
                        cut_left = np.nanpercentile(norm_par_array, 100 - percentile)
                        cut_right = np.nanpercentile(norm_par_array, percentile)

                    elif isinstance(percentile, dict):
                        if "low_side" in percentile:
                            cut_left = np.nanpercentile(norm_par_array, percentile)
                        else:
                            cut_left = None
                        if "high_side" in percentile:
                            cut_right = np.nanpercentile(norm_par_array, percentile)
                        else:
                            cut_right = None

            if default is not None:
                value = default["value"]
                default_mode = default["mode"]
                if isinstance(value, (int, float)):
                    default_cut_left = -value
                    default_cut_right = value
                else:
                    if "low_side" in default:
                        default_cut_left = value["low_side"]
                    else:
                        default_cut_left = np.nan
                    if "high_side" in default:
                        default_cut_right = value["high_side"]
                    else:
                        default_cut_right = np.nan

                if default_mode == "higher_limit":
                    if cut_left is not None:
                        if cut_left < default_cut_left:
                            cut_left = default_cut_left
                    if cut_right is not None:
                        if cut_right > default_cut_right:
                            cut_right = default_cut_right
                elif default_mode == "lower_limit":
                    if cut_left is not None:
                        if cut_left > default_cut_left:
                            cut_left = default_cut_left
                    if cut_right is not None:
                        if cut_right < default_cut_right:
                            cut_right = default_cut_right
                else:
                    raise ValueError("unknown mode")

            if mode == "inclusive":
                if cut_right is not None and cut_left is not None:
                    cut_string = f"({out_par}_classifier>a) & ({out_par}_classifier<b)"
                    par_dict = {"a": cut_left, "b": cut_right}
                elif cut_right is None:
                    cut_string = f"{out_par}_classifier>a"
                    par_dict = {"a": cut_left}
                elif cut_left is None:
                    cut_string = f"{out_par}_classifier<a"
                    par_dict = {"a": cut_right}
            elif mode == "exclusive":
                if cut_right is not None and cut_left is not None:
                    cut_string = f"({out_par}_classifier<a) | ({out_par}_classifier>b)"
                    par_dict = {"a": cut_left, "b": cut_right}
                elif cut_right is None:
                    cut_string = f"{out_par}_classifier<a"
                    par_dict = {"a": cut_left}
                elif cut_left is None:
                    cut_string = f"{out_par}_classifier>a"
                    par_dict = {"a": cut_right}

            output_dict[f"{out_par}_classifier"] = {
                "expression": f"(({par})-a)/b",
                "parameters": {"a": mean, "b": std},
            }

            output_dict[out_par] = {"expression": cut_string, "parameters": par_dict}
            if display > 0:
                fig = plt.figure()
                low = -10 if cut_left is None or cut_left > -10 else cut_left
                hi = 10 if cut_right is None or cut_right < 10 else cut_right
                hist, _, _ = plt.hist(
                    norm_par_array,
                    bins=np.arange(low, hi, 0.1),
                    histtype="step",
                )
                if percentile is not None and method == "fit":
                    xs = np.arange(low, hi, 0.1)
                    if func == skewed_fit:
                        pdf_values = func(xs, *pars)[1] * 0.1
                    else:
                        pdf_values = func.pdf_ext(xs, *pars)[1] * 0.1
                    plt.plot(xs, pdf_values)
                if cut_left is not None:
                    plt.axvline(cut_left)
                if cut_right is not None:
                    plt.axvline(cut_right)

                plt.ylabel("counts")
                plt.xlabel(f"{out_par}_classifier")
                plot_dict[out_par] = fig
                plt.close()
    if display > 0:
        return output_dict, plot_dict
    else:
        return output_dict


def find_pulser_properties(df, energy="daqenergy"):
    """
    Searches for pulser in the energy spectrum using time between events in peaks
    """
    if np.nanmax(df[energy]) > 8000:
        hist, bins, var = pgh.get_hist(
            df[energy], dx=1, range=(1000, np.nanmax(df[energy]))
        )
        allowed_err = 200
    else:
        hist, bins, var = pgh.get_hist(
            df[energy], dx=0.2, range=(500, np.nanmax(df[energy]))
        )
        allowed_err = 50
    if np.any(var == 0):
        var[np.where(var == 0)] = 1
    imaxes = pgc.get_i_local_maxima(hist / np.sqrt(var), 3)
    peak_energies = pgh.get_bin_centers(bins)[imaxes]
    pt_pars, pt_covs = pgc.hpge_fit_E_peak_tops(
        hist, bins, var, peak_energies, n_to_fit=10
    )
    peak_e_err = pt_pars[:, 1] * 4

    allowed_mask = np.ones(len(peak_energies), dtype=bool)
    for i, e in enumerate(peak_energies[1:-1]):
        i += 1
        if peak_e_err[i] > allowed_err:
            continue
        if i == 1:
            if (
                e - peak_e_err[i] < peak_energies[i - 1] + peak_e_err[i - 1]
                and peak_e_err[i - 1] < allowed_err
            ):
                overlap = (
                    peak_energies[i - 1]
                    + peak_e_err[i - 1]
                    - (peak_energies[i] - peak_e_err[i])
                )
                peak_e_err[i] -= overlap * (
                    peak_e_err[i] / (peak_e_err[i] + peak_e_err[i - 1])
                )
                peak_e_err[i - 1] -= overlap * (
                    peak_e_err[i - 1] / (peak_e_err[i] + peak_e_err[i - 1])
                )

        if (
            e + peak_e_err[i] > peak_energies[i + 1] - peak_e_err[i + 1]
            and peak_e_err[i + 1] < allowed_err
        ):
            overlap = (e + peak_e_err[i]) - (peak_energies[i + 1] - peak_e_err[i + 1])
            total = peak_e_err[i] + peak_e_err[i + 1]
            peak_e_err[i] -= (overlap) * (peak_e_err[i] / total)
            peak_e_err[i + 1] -= (overlap) * (peak_e_err[i + 1] / total)

    out_pulsers = []
    for i, e in enumerate(peak_energies[allowed_mask]):
        if peak_e_err[i] > allowed_err:
            continue

        try:
            e_cut = (df[energy] > e - peak_e_err[i]) & (df[energy] < e + peak_e_err[i])
            df_peak = df[e_cut]

            time_since_last = (
                df_peak.timestamp.values[1:] - df_peak.timestamp.values[:-1]
            )

            tsl = time_since_last[
                (time_since_last >= 0)
                & (time_since_last < np.percentile(time_since_last, 99.9))
            ]

            bins = np.arange(0.1, 5, 0.001)
            bcs = pgh.get_bin_centers(bins)
            hist, bins, var = pgh.get_hist(tsl, bins=bins)

            maxs = pgh.get_i_local_maxima(hist, 45)
            maxs = maxs[maxs > 20]

            super_max = pgh.get_i_local_maxima(hist, 500)
            super_max = super_max[super_max > 20]
            if len(maxs) < 2:
                continue
            else:
                max_locs = np.array([0.0])
                max_locs = np.append(max_locs, bcs[np.array(maxs)])
                if (
                    len(np.where(np.abs(np.diff(np.diff(max_locs))) <= 0.001)[0]) > 1
                    or (np.abs(np.diff(np.diff(max_locs))) <= 0.001).all()
                    or len(super_max) > 0
                ):
                    pulser_e = e
                    period = stats.mode(tsl).mode[0]
                    if period > 0.1:
                        out_pulsers.append((pulser_e, peak_e_err[i], period, energy))

                else:
                    continue
        except Exception:
            continue
    return out_pulsers


def get_tcm_pulser_ids(tcm_file, channel, multiplicity_threshold):
    if isinstance(channel, str):
        if channel[:2] == "ch":
            chan = int(channel[2:])
        else:
            chan = int(channel)
    else:
        chan = channel
    if isinstance(tcm_file, list):
        mask = np.array([], dtype=bool)
        for file in tcm_file:
            _, file_mask = get_tcm_pulser_ids(file, chan, multiplicity_threshold)
            mask = np.append(mask, file_mask)
        ids = np.where(mask)[0]
    else:
        array_ids = lh5.read("hardware_tcm_1/table_key", tcm_file).view_as("ak")
        chan_evts = ak.any(array_ids == chan, axis=-1)
        multiplicity = ak.count(array_ids[chan_evts], axis=-1)
        ids = np.where(multiplicity > multiplicity_threshold)[0]
        mask = np.zeros(len(array_ids[chan_evts]), dtype=bool)
        mask[ids] = True
    return ids, mask


def tag_pulsers(df, chan_info, window=0.01):
    df["isPulser"] = 0

    if isinstance(chan_info, tuple):
        chan_info = [chan_info]
    final_mask = None
    for chan_i in chan_info:
        pulser_energy, peak_e_err, period, energy_name = chan_i

        e_cut = (df[energy_name] < pulser_energy + peak_e_err) & (
            df[energy_name] > pulser_energy - peak_e_err
        )
        df_pulser = df[e_cut]

        time_since_last = np.zeros(len(df_pulser))
        time_since_last[1:] = (
            df_pulser.timestamp.values[1:] - df_pulser.timestamp.values[:-1]
        )

        mode_idxs = (time_since_last > period - window) & (
            time_since_last < period + window
        )

        pulser_events = np.count_nonzero(mode_idxs)
        # print(f"pulser events: {pulser_events}")
        if pulser_events < 3:
            return df
        df_pulser = df_pulser[mode_idxs]

        ts = df_pulser.timestamp.values
        diff_zero = np.zeros(len(ts))
        diff_zero[1:] = np.around(np.divide(np.subtract(ts[1:], ts[:-1]), period))
        diff_cum = np.cumsum(diff_zero)
        z = np.polyfit(diff_cum, ts, 1)

        period = z[0]
        phase = z[1]
        mod = np.abs(df.timestamp - phase) % period

        period_cut = (mod < 0.1) | ((period - mod) < 0.1)  # 0.1)

        if final_mask is None:
            final_mask = e_cut & period_cut
        else:
            final_mask = final_mask | (e_cut & period_cut)

    df.loc[final_mask, "isPulser"] = 1

    return df



"""
This module contains the functions for performing the energy optimisation.
This happens in 2 steps, firstly a grid search is performed on each peak
separately using the optimiser, then the resulting grids are interpolated
to provide the best energy resolution at Qbb
"""

import logging

import lgdo.lh5 as lh5
import matplotlib.pyplot as plt
import numpy as np

import pygama.math.distributions as pgd
import pygama.math.histogram as pgh
import pygama.pargen.energy_cal as pgc
from pygama.pargen.utils import convert_to_minuit, return_nans

log = logging.getLogger(__name__)
sto = lh5.LH5Store()


def simple_guess(energy, func, fit_range=None, bin_width=None):
    """
    Simple guess for peak fitting
    """
    if fit_range is None:
        fit_range = (np.nanmin(energy), np.nanmax(energy))

    energy = energy[(energy >= fit_range[0]) & (energy <= fit_range[1])]
    if bin_width is None:
        init_bin_width = (
            2
            * (np.nanpercentile(energy, 75) - np.nanpercentile(energy, 25))
            * len(energy) ** (-1 / 3)
        )
        init_hist, init_bins, _ = pgh.get_hist(
            energy, dx=init_bin_width, range=fit_range
        )
        try:
            _, init_sigma, _ = pgh.get_gaussian_guess(init_hist, init_bins)
        except IndexError:
            init_hist, init_bins, _ = pgh.get_hist(
                energy, dx=init_bin_width / 2, range=fit_range
            )
            try:
                _, init_sigma, _ = pgh.get_gaussian_guess(init_hist, init_bins)
            except IndexError:
                init_sigma = np.nanstd(energy)
        bin_width = (init_sigma) * len(energy) ** (-1 / 3)

    hist, bins, var = pgh.get_hist(energy, dx=bin_width, range=fit_range)

    # make binning dynamic based on max, % of events/ n of events?
    hist, bins, var = pgh.get_hist(energy, range=fit_range, dx=bin_width)

    if func == pgd.hpge_peak or func == pgd.gauss_on_step:
        mu, sigma, amp = pgh.get_gaussian_guess(hist, bins)
        i_0 = np.argmax(hist)
        bg = np.mean(hist[-10:])
        step = bg - np.mean(hist[:10])
        hstep = step / (bg + np.mean(hist[:10]))
        dx = np.diff(bins)[0]
        n_bins_range = int((4 * sigma) // dx)
        nsig = np.sum(hist[i_0 - n_bins_range : i_0 + n_bins_range])
        nbkg = np.sum(hist) - nsig

        parguess = {
            "n_sig": nsig,
            "mu": mu,
            "sigma": sigma,
            "n_bkg": nbkg,
            "hstep": hstep,
            "x_lo": fit_range[0],
            "x_hi": fit_range[1],
        }

        if func == pgd.hpge_peak:
            htail = 1.0 / 5
            tau = 0.5 * sigma
            parguess["htail"] = htail
            parguess["tau"] = tau

    else:
        log.error(f"simple_guess not implemented for {func.__name__}")
        return return_nans(func)

    return convert_to_minuit(parguess, func).values


def get_peak_fwhm_with_dt_corr(
    energies,
    alpha,
    dt,
    func,
    peak,
    kev_width,
    guess=None,
    kev=False,
    frac_max=0.5,
    bin_width=1,
    allow_tail_drop=False,
    display=0,
):
    """
    Applies the drift time correction and fits the peak returns the fwhm, fwhm/max and associated errors,
    along with the number of signal events and the reduced chi square of the fit. Can return result in ADC or keV.
    """

    correction = np.multiply(
        np.multiply(alpha, dt, dtype="float64"), energies, dtype="float64"
    )
    ct_energy = np.add(correction, energies)

    lower_bound = (np.nanmin(ct_energy) // bin_width) * bin_width
    upper_bound = ((np.nanmax(ct_energy) // bin_width) + 1) * bin_width
    hist, bins, var = pgh.get_hist(
        ct_energy, dx=bin_width, range=(lower_bound, upper_bound)
    )
    mu = bins[np.nanargmax(hist)]
    adc_to_kev = mu / peak
    # Making the window slightly smaller removes effects where as mu moves edge can be outside bin width
    lower_bound = mu - ((kev_width[0] - 2) * adc_to_kev)
    upper_bound = mu + ((kev_width[1] - 2) * adc_to_kev)
    win_idxs = (ct_energy > lower_bound) & (ct_energy < upper_bound)
    fit_range = (lower_bound, upper_bound)
    tol = None
    try:
        (
            energy_pars,
            energy_err,
            cov,
            chisqr,
            func,
            _,
            _,
            _,
        ) = pgc.unbinned_staged_energy_fit(
            ct_energy[win_idxs],
            func=func,
            fit_range=fit_range,
            guess_func=simple_guess,
            tol=tol,
            guess=guess,
            allow_tail_drop=allow_tail_drop,
            bin_width=bin_width,
            display=display,
        )
        if display > 0:
            plt.figure()
            xs = np.arange(lower_bound, upper_bound, bin_width)
            fit_hist, fit_bins, _ = pgh.get_hist(
                ct_energy, dx=bin_width, range=(lower_bound, upper_bound)
            )
            plt.step(pgh.get_bin_centers(fit_bins), fit_hist)
            plt.plot(xs, func.get_pdf(xs, *energy_pars))
            plt.show()

        fwhm = func.get_fwfm(energy_pars, frac_max=frac_max)

        xs = np.arange(lower_bound, upper_bound, 0.1)
        y = func.get_pdf(xs, *energy_pars)
        max_val = np.amax(y)
        fwhm_o_max = fwhm / max_val

        rng = np.random.default_rng(1)
        # generate set of bootstrapped parameters
        par_b = rng.multivariate_normal(energy_pars, cov, size=100)
        y_max = np.array([func.get_pdf(xs, *p) for p in par_b])
        maxs = np.nanmax(y_max, axis=1)

        y_b = np.zeros(len(par_b))
        for i, p in enumerate(par_b):
            try:
                y_b[i] = func.get_fwfm(p, frac_max=frac_max)
            except Exception:
                y_b[i] = np.nan
        fwhm_err = np.nanstd(y_b, axis=0)
        fwhm_o_max_err = np.nanstd(y_b / maxs, axis=0)

        if display > 1:
            plt.figure()
            plt.step(pgh.get_bin_centers(bins), hist)
            for i in range(100):
                plt.plot(xs, y_max[i, :])
            plt.show()

        if display > 0:
            plt.figure()
            hist, bins, var = pgh.get_hist(
                ct_energy, dx=bin_width, range=(lower_bound, upper_bound)
            )
            plt.step(pgh.get_bin_centers(bins), hist)
            plt.plot(xs, y, color="orange")
            yerr_boot = np.nanstd(y_max, axis=0)
            plt.fill_between(
                xs, y - yerr_boot, y + yerr_boot, facecolor="C1", alpha=0.5
            )
            plt.show()

    except Exception:
        return np.nan, np.nan, np.nan, np.nan, (np.nan, np.nan), np.nan, np.nan, None

    if kev is True:
        fwhm *= peak / energy_pars["mu"]
        fwhm_err *= peak / energy_pars["mu"]

    return (
        fwhm,
        fwhm_o_max,
        fwhm_err,
        fwhm_o_max_err,
        chisqr,
        energy_pars["n_sig"],
        energy_err["n_sig"],
        energy_pars,
    )


def fom_fwhm_with_alpha_fit(
    tb_in, kwarg_dict, ctc_parameter, nsteps=11, idxs=None, frac_max=0.2, display=0
):
    """
    FOM for sweeping over ctc values to find the best value, returns the best found fwhm with its error,
    the corresponding alpha value and the number of events in the fitted peak, also the reduced chisquare of the
    """
    parameter = kwarg_dict["parameter"]
    func = kwarg_dict["func"]
    energies = tb_in[parameter].nda
    energies = energies.astype("float64")
    peak = kwarg_dict["peak"]
    kev_width = kwarg_dict["kev_width"]
    bin_width = kwarg_dict.get("bin_width", 1)
    min_alpha = 0
    max_alpha = 3.50e-06
    alphas = np.linspace(min_alpha, max_alpha, nsteps, dtype="float64")
    try:
        dt = tb_in[ctc_parameter].nda
    except KeyError:
        dt = tb_in.eval(ctc_parameter)
    if idxs is not None:
        energies = energies[idxs]
        dt = dt[idxs]
    try:
        if np.isnan(energies).any():
            log.debug("nan in energies")
            raise RuntimeError
        if np.isnan(dt).any():
            log.debug("nan in dts")
            raise RuntimeError
        fwhms = np.array([])
        final_alphas = np.array([])
        fwhm_errs = np.array([])
        best_fwhm = np.inf
        early_break = False
        for alpha in alphas:
            (
                _,
                fwhm_o_max,
                _,
                fwhm_o_max_err,
                _,
                _,
                _,
                fit_pars,
            ) = get_peak_fwhm_with_dt_corr(
                energies,
                alpha,
                dt,
                func,
                peak,
                kev_width,
                guess=None,
                frac_max=0.5,
                bin_width=bin_width,
                allow_tail_drop=False,
            )
            if not np.isnan(fwhm_o_max):
                fwhms = np.append(fwhms, fwhm_o_max)
                final_alphas = np.append(final_alphas, alpha)
                fwhm_errs = np.append(fwhm_errs, fwhm_o_max_err)
                if fwhms[-1] < best_fwhm:
                    best_fwhm = fwhms[-1]
            log.info(f"alpha: {alpha}, fwhm/max:{fwhm_o_max:.4f}+-{fwhm_o_max_err:.4f}")

            ids = (fwhm_errs < 2 * np.nanpercentile(fwhm_errs, 50)) & (
                fwhm_errs > 1e-10
            )
            if len(fwhms[ids]) > 5:
                if (np.diff(fwhms[ids])[-3:] > 0).all():
                    early_break = True
                    break

        # Make sure fit isn't based on only a few points
        if len(fwhms) < nsteps * 0.2 and early_break is False:
            log.debug("less than 20% fits successful")
            raise RuntimeError

        ids = (fwhm_errs < 2 * np.nanpercentile(fwhm_errs, 50)) & (fwhm_errs > 1e-10)
        # Fit alpha curve to get best alpha

        try:
            alphas = np.linspace(
                final_alphas[ids][0],
                final_alphas[ids][-1],
                nsteps * 20,
                dtype="float64",
            )
            alpha_fit, cov = np.polyfit(
                final_alphas[ids], fwhms[ids], w=1 / fwhm_errs[ids], deg=4, cov=True
            )
            fit_vals = np.polynomial.polynomial.polyval(alphas, alpha_fit[::-1])
            alpha = alphas[np.nanargmin(fit_vals)]

            rng = np.random.default_rng(1)
            alpha_pars_b = rng.multivariate_normal(alpha_fit, cov, size=1000)
            fits = np.array(
                [
                    np.polynomial.polynomial.polyval(alphas, pars[::-1])
                    for pars in alpha_pars_b
                ]
            )
            min_alphas = np.array([alphas[np.nanargmin(fit)] for fit in fits])
            alpha_err = np.nanstd(min_alphas)
            if display > 0:
                plt.figure()
                yerr_boot = np.nanstd(fits, axis=0)
                plt.errorbar(final_alphas, fwhms, yerr=fwhm_errs, linestyle=" ")
                plt.plot(alphas, fit_vals)
                plt.fill_between(
                    alphas,
                    fit_vals - yerr_boot,
                    fit_vals + yerr_boot,
                    facecolor="C1",
                    alpha=0.5,
                )
                plt.show()

        except Exception:
            log.debug("alpha fit failed")

        if np.isnan(fit_vals).all():
            log.debug("alpha fit all nan")
            raise RuntimeError
        (
            final_fwhm,
            _,
            final_err,
            _,
            csqr,
            n_sig,
            n_sig_err,
            _,
        ) = get_peak_fwhm_with_dt_corr(
            energies,
            alpha,
            dt,
            func,
            peak,
            kev_width,
            guess=None,
            kev=True,
            frac_max=frac_max,
            allow_tail_drop=True,
            bin_width=bin_width,
            display=display,
        )
        if np.isnan(final_fwhm) or np.isnan(final_err):
            log.debug(f"final fit failed, alpha was {alpha}")
            raise RuntimeError
        return {
            "fwhm": final_fwhm,
            "fwhm_err": final_err,
            "alpha": alpha,
            "alpha_err": alpha_err,
            "chisquare": csqr,
            "n_sig": n_sig,
            "n_sig_err": n_sig_err,
        }
    except Exception:
        return {
            "fwhm": np.nan,
            "fwhm_err": np.nan,
            "alpha": 0,
            "alpha_err": np.nan,
            "chisquare": (np.nan, np.nan),
            "n_sig": np.nan,
            "n_sig_err": np.nan,
        }


def fom_fwhm_no_alpha_sweep(
    tb_in, kwarg_dict, ctc_param=None, alpha=0, idxs=None, frac_max=0.5, display=0
):
    """
    FOM with no ctc sweep, used for optimising ftp.
    """
    parameter = kwarg_dict["parameter"]
    func = kwarg_dict["func"]
    energies = tb_in[parameter].nda
    energies = energies.astype("float64")
    peak = kwarg_dict["peak"]
    kev_width = kwarg_dict["kev_width"]
    alpha = kwarg_dict.get("alpha", alpha)
    bin_width = kwarg_dict.get("bin_width", 1)
    if isinstance(alpha, dict):
        alpha = alpha[parameter]
    if "ctc_param" in kwarg_dict or ctc_param is not None:
        ctc_param = kwarg_dict.get("ctc_param", ctc_param)
        try:
            dt = tb_in[ctc_param].nda
        except KeyError:
            dt = tb_in.eval(ctc_param)
            dt = tb_in[ctc_param].nda
    else:
        dt = 0

    if idxs is not None:
        energies = energies[idxs]
        dt = dt[idxs]

    if np.isnan(energies).any():
        return {
            "fwhm": np.nan,
            "fwhm_o_max": np.nan,
            "fwhm_err": np.nan,
            "fwhm_o_max_err": np.nan,
            "chisquare": np.nan,
            "n_sig": np.nan,
            "n_sig_err": np.nan,
        }
    (
        fwhm,
        final_fwhm_o_max,
        fwhm_err,
        final_fwhm_o_max_err,
        csqr,
        n_sig,
        n_sig_err,
        fit_pars,
    ) = get_peak_fwhm_with_dt_corr(
        energies,
        alpha,
        dt,
        func,
        peak=peak,
        kev_width=kev_width,
        frac_max=frac_max,
        kev=True,
        bin_width=bin_width,
        display=display,
    )
    return {
        "fwhm": fwhm,
        "fwhm_o_max": final_fwhm_o_max,
        "fwhm_err": fwhm_err,
        "fwhm_o_max_err": final_fwhm_o_max_err,
        "chisquare": csqr,
        "n_sig": n_sig,
        "n_sig_err": n_sig_err,
    }


def fom_single_peak_alpha_sweep(data, kwarg_dict, display=0):
    idx_list = kwarg_dict["idx_list"]
    ctc_param = kwarg_dict["ctc_param"]
    peak_dicts = kwarg_dict["peak_dicts"]
    frac_max = kwarg_dict.get("frac_max", 0.2)
    out_dict = fom_fwhm_with_alpha_fit(
        data,
        peak_dicts[0],
        ctc_param,
        idxs=idx_list[0],
        frac_max=frac_max,
        display=display,
    )
    return out_dict


def fom_interpolate_energy_res_with_single_peak_alpha_sweep(
    data, kwarg_dict, display=0
):
    peaks = kwarg_dict["peaks_kev"]
    idx_list = kwarg_dict["idx_list"]
    ctc_param = kwarg_dict["ctc_param"]
    peak_dicts = kwarg_dict["peak_dicts"]
    interp_energy = kwarg_dict.get("interp_energy", {"Qbb": 2039})
    fwhm_func = kwarg_dict.get("fwhm_func", pgc.FWHMLinear)
    frac_max = kwarg_dict.get("frac_max", 0.2)

    out_dict = fom_fwhm_with_alpha_fit(
        data,
        peak_dicts[-1],
        ctc_param,
        idxs=idx_list[-1],
        frac_max=frac_max,
        display=display,
    )
    alpha = out_dict["alpha"]
    log.info(alpha)
    fwhms = []
    fwhm_errs = []
    n_sig = []
    n_sig_err = []
    for i, _ in enumerate(peaks[:-1]):
        out_peak_dict = fom_fwhm_no_alpha_sweep(
            data,
            peak_dicts[i],
            ctc_param,
            alpha=alpha,
            idxs=idx_list[i],
            frac_max=frac_max,
            display=display,
        )
        fwhms.append(out_peak_dict["fwhm"])
        fwhm_errs.append(out_peak_dict["fwhm_err"])
        n_sig.append(out_peak_dict["n_sig"])
        n_sig_err.append(out_peak_dict["n_sig_err"])
    fwhms.append(out_dict["fwhm"])
    fwhm_errs.append(out_dict["fwhm_err"])
    n_sig.append(out_dict["n_sig"])
    n_sig_err.append(out_dict["n_sig_err"])
    log.info(f"fwhms are {fwhms}keV +- {fwhm_errs}")

    fwhms = np.array(fwhms)
    fwhm_errs = np.array(fwhm_errs)
    n_sig = np.array(n_sig)
    n_sig_err = np.array(n_sig_err)
    peaks = np.array(peaks)

    nan_mask = np.isnan(fwhms) | (fwhms < 0)
    if len(fwhms[~nan_mask]) < 2:
        return np.nan, np.nan, np.nan
    else:
        results = pgc.HPGeCalibration.fit_energy_res_curve(
            fwhm_func, peaks[~nan_mask], fwhms[~nan_mask], fwhm_errs[~nan_mask]
        )
        results = pgc.HPGeCalibration.interpolate_energy_res(
            fwhm_func, peaks[~nan_mask], results, interp_energy
        )
        interp_res = results[f"{list(interp_energy)[0]}_fwhm_in_kev"]
        interp_res_err = results[f"{list(interp_energy)[0]}_fwhm_err_in_kev"]

        if nan_mask[-1] is True or nan_mask[-2] is True:
            interp_res_err = np.nan
        if interp_res_err / interp_res > 0.1:
            interp_res_err = np.nan

    log.info(f"{list(interp_energy)[0]} fwhm is {interp_res} keV +- {interp_res_err}")

    return {
        f"{list(interp_energy)[0]}_fwhm": interp_res,
        f"{list(interp_energy)[0]}_fwhm_err": interp_res_err,
        "alpha": alpha,
        "peaks": peaks.tolist(),
        "fwhms": fwhms,
        "fwhm_errs": fwhm_errs,
        "n_sig": n_sig,
        "n_sig_err": n_sig_err,
    }



"""
Step distributions for pygama
"""

from math import erf

import numba as nb
import numpy as np

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_defaults(parallel=False))
def nb_step_int(x: float, mu: float, sigma: float, hstep: float) -> np.ndarray:
    r"""
    Integral of step function w/args mu, sigma, hstep. It computes:


    .. math::
        \int cdf(x, hstep, \mu, \sigma)\, dx = \sigma\left(\frac{x-\mu}{\sigma} + hstep \left(\frac{x-\mu}{\sigma}\mathrm{erf}\left(\frac{x-\mu}{\sigma\sqrt{2}}\right) + \sqrt{\frac{2}{\pi}}\exp\left(-(\frac{x-\mu}{\sigma})^2/2\right) \right)\right)


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        A single input data point
    mu
        The location of the step
    sigma
        The width of the step
    hstep
        The height of the step


    Returns
    -------
    step_int
        The cumulative integral of the step distribution at x
    """

    if sigma == 0:
        y = (x) + hstep * (x - mu)
    else:
        z = (x - mu) / sigma
        y = sigma * (
            z
            + hstep
            * (z * erf(z / np.sqrt(2)) + np.sqrt(2 / np.pi) * np.exp(-(z**2) / 2))
        )
    return y


@nb.njit(**nb_defaults(parallel=False))
def nb_unnorm_step_pdf(x: float, mu: float, sigma: float, hstep: float) -> float:
    r"""
    Unnormalised step function for use in pdfs. It computes:


    .. math::
        pdf(x, hstep, \mu, \sigma) = 1+ hstep\mathrm{erf}\left(\frac{x-\mu}{\sigma\sqrt{2}}\right)



    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        A single input data point
    mu
        The location of the step
    sigma
        The "width" of the step, because we are using an error function to define it
    hstep
        The height of the step

    """

    invs = np.sqrt(2) * sigma
    if invs == 0:
        return 1 + hstep
    else:
        step_f = 1 + hstep * erf((x - mu) / invs)
        return step_f


@nb.njit(**nb_kwargs)
def nb_step_pdf(
    x: np.ndarray, x_lo: float, x_hi: float, mu: float, sigma: float, hstep: float
) -> np.ndarray:
    r"""
    Normalised step function w/args mu, sigma, hstep, x_lo, x_hi. Its range of support is :math:`x\in` (x_lo, x_hi). It computes:


    .. math::
        pdf(x, \mathrm{x}_\mathrm{lo}, \mathrm{x}_\mathrm{hi}, \mu, \sigma, hstep) = pdf(y=\frac{x-\mu}{\sigma}, step, \mathrm{x}_\mathrm{lo}, \mathrm{x}_\mathrm{hi}) = \frac{1+hstep\mathrm{erf}\left(\frac{x-\mu}{\sigma\sqrt{2}}\right)}{\sigma\left[(y-y_{min}) +hstep\left(y\mathrm{erf}(\frac{y}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y^2/2}-y_{min}\mathrm{erf}(\frac{y_{min}}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y_{min}^2/2}\right)\right]}


    Where :math:`y_{max} = \frac{\mathrm{x}_\mathrm{hi} - \mu}{\sigma}, y_{min} = \frac{\mathrm{x}_\mathrm{lo} - \mu}{\sigma}`.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        The input data
    x_lo
        The lower range on which to normalize the step PDF, default is to normalize from min to max x values
    x_hi
        The upper range on which to normalize the step PDF
    mu
        The location of the step
    sigma
        The "width" of the step, because we are using an error function to define it
    hstep
        The height of the step
    """
    # Compute the normalization
    norm = nb_step_int(x_hi, mu, sigma, hstep) - nb_step_int(x_lo, mu, sigma, hstep)

    if norm == 0:
        # If the normalization is zero, don't waste time computing the step_pdf
        z = np.full_like(x, np.inf, dtype=np.float64)

    else:
        z = np.empty_like(x, dtype=np.float64)
        for i in nb.prange(x.shape[0]):
            step_f = nb_unnorm_step_pdf(x[i], mu, sigma, hstep)
            z[i] = step_f / norm

    return z


@nb.njit(**nb_kwargs)
def nb_step_cdf(
    x: np.ndarray, x_lo: float, x_hi: float, mu: float, sigma: float, hstep: float
) -> np.ndarray:
    r"""
    Normalized CDF for step function w/args mu, sigma, hstep, x_lo, x_hi. Its range of support is :math:`x\in` (x_lo, x_hi). It computes:


    .. math::
        cdf(x, \mathrm{x}_\mathrm{lo}, \mathrm{x}_\mathrm{hi}, \mu, \sigma, hstep) = cdf(y=\frac{x-\mu}{\sigma}, hstep, \mathrm{x}_\mathrm{lo}, \mathrm{x}_\mathrm{hi}) = \frac{(y-y_{min}) +hstep\left(y\mathrm{erf}(\frac{y}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y^2/2}-y_{min}\mathrm{erf}(\frac{y_{min}}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y_{min}^2/2}\right)}{\sigma\left[(y_{max}-y_{min}) +hstep\left(y_{max}\mathrm{erf}(\frac{y_{max}}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y_{max}^2/2}-y_{min}\mathrm{erf}(\frac{y_{min}}{\sqrt{2}})+\sqrt{\frac{2}{\pi}}e^{-y_{min}^2/2}\right)\right] }


    Where :math:`y_{max} = \frac{\mathrm{x}_\mathrm{hi} - \mu}{\sigma}, y_{min} = \frac{\mathrm{x}_\mathrm{lo} - \mu}{\sigma}`.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower range on which to normalize the step PDF, default is to normalize from min to max x values
    x_hi
        The upper range on which to normalize the step PDF
    mu
        The location of the step
    sigma
        The "width" of the step, because we are using an error function to define it
    hstep
        The height of the step
    """
    integral_lo = nb_step_int(x_lo, mu, sigma, hstep)
    integral_hi = nb_step_int(x_hi, mu, sigma, hstep)
    norm = integral_hi - integral_lo

    if norm == 0:
        # If the normalization is zero, return np.inf and avoid wasting time computing the integral
        cdf = np.full_like(x, np.inf, dtype=np.float64)

    else:
        integrated_pdf = np.empty_like(x, dtype=np.float64)

        for i in nb.prange(x.shape[0]):
            integrated_pdf[i] = nb_step_int(x[i], mu, sigma, hstep)
        cdf = (integrated_pdf - integral_lo) / norm  # compute (cdf(x)-cdf(x_min))/norm

    return cdf


@nb.njit(**nb_defaults(parallel=False))
def nb_step_scaled_pdf(
    x: np.ndarray,
    x_lo: float,
    x_hi: float,
    area: float,
    mu: float,
    sigma: float,
    hstep: float,
) -> np.ndarray:
    r"""
    Scaled step function pdf w/args x_lo, x_hi, area, mu, sigma, hstep.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower range on which to normalize the step PDF, default is to normalize from min to max x values
    x_hi
        The upper range on which to normalize the step PDF
    area
        The number to scale the distribution by
    mu
        The location of the step
    sigma
        The "width" of the step, because we are using an error function to define it
    hstep
        The height of the step
    """

    return area * nb_step_pdf(x, x_lo, x_hi, mu, sigma, hstep)


@nb.njit(**nb_defaults(parallel=False))
def nb_step_scaled_cdf(
    x: np.ndarray,
    x_lo: float,
    x_hi: float,
    area: float,
    mu: float,
    sigma: float,
    hstep: float,
) -> np.ndarray:
    r"""
    Scaled step function CDF w/args  x_lo, x_hi, area, mu, sigma, hstep. Used for extended binned fits.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    x_lo
        The lower range on which to normalize the step PDF, default is to normalize from min to max x values
    x_hi
        The upper range on which to normalize the step PDF
    area
        The number to scale the distribution by
    mu
        The location of the step
    sigma
        The "width" of the step, because we are using an error function to define it
    hstep
        The height of the step
    """

    return area * nb_step_cdf(x, x_lo, x_hi, mu, sigma, hstep)


class StepGen(PygamaContinuous):

    def _argcheck(self, x_lo, x_hi, mu, sigma, hstep):
        return x_hi > x_lo

    def __init__(self, *args, **kwargs):
        self.x_lo = None
        self.x_hi = None
        super().__init__(*args, **kwargs)

    def _pdf(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        x.flags.writeable = True
        return nb_step_pdf(x, x_lo[0], x_hi[0], mu[0], sigma[0], hstep[0])

    def _cdf(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        x.flags.writeable = True
        return nb_step_cdf(x, x_lo[0], x_hi[0], mu[0], sigma[0], hstep[0])

    def get_pdf(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return nb_step_pdf(x, x_lo, x_hi, mu, sigma, hstep)

    def get_cdf(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return nb_step_cdf(x, x_lo, x_hi, mu, sigma, hstep)

    # Because step is only defined on a user specified range, we don't need to return a different pdf_norm, just alias get_pdf and get_cdf
    def pdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return nb_step_pdf(x, x_lo, x_hi, mu, sigma, hstep)

    def cdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return nb_step_cdf(x, x_lo, x_hi, mu, sigma, hstep)

    def pdf_ext(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        area: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return np.diff(
            nb_step_scaled_cdf(
                np.array([x_lo, x_hi]), x_lo, x_hi, area, mu, sigma, hstep
            )
        )[0], nb_step_scaled_pdf(x, x_lo, x_hi, area, mu, sigma, hstep)

    def cdf_ext(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        area: float,
        mu: float,
        sigma: float,
        hstep: float,
    ) -> np.ndarray:
        return nb_step_scaled_cdf(x, x_lo, x_hi, area, mu, sigma, hstep)

    def required_args(self) -> tuple[str, str, str, str, str]:
        return "x_lo", "x_hi", "mu", "sigma", "hstep"


step = StepGen(name="step")



import os
from pathlib import Path

import pytest
from dspeed import build_dsp
from legendtestdata import LegendTestData

config_dir = Path(__file__).parent / "configs"


@pytest.fixture(scope="session")
def lgnd_test_data():
    ldata = LegendTestData()
    ldata.checkout("756aef8")
    return ldata


@pytest.fixture(scope="session")
def tmp_dir(tmpdir_factory):
    out_dir = tmpdir_factory.mktemp("data")
    assert os.path.exists(out_dir)
    return out_dir


@pytest.fixture(scope="session")
def dsp_test_file(lgnd_test_data, tmp_dir):
    out_name = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_dsp.lh5"
    build_dsp(
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        out_name,
        dsp_config=f"{config_dir}/icpc-dsp-config.json",
        database={"pz": {"tau": 27460.5}},
        write_mode="r",
    )
    assert os.path.exists(out_name)

    return out_name


@pytest.fixture(scope="session")
def dsp_test_file_spm(lgnd_test_data, tmp_dir):
    chan_config = {
        "ch0/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch1/raw": f"{config_dir}/sipm-dsp-config.json",
        "ch2/raw": f"{config_dir}/sipm-dsp-config.json",
    }

    out_file = f"{tmp_dir}/L200-comm-20211130-phy-spms_dsp.lh5"
    build_dsp(
        lgnd_test_data.get_path("lh5/L200-comm-20211130-phy-spms.lh5"),
        out_file,
        {},
        n_max=5,
        lh5_tables=chan_config.keys(),
        chan_config=chan_config,
        write_mode="r",
    )

    assert os.path.exists(out_file)

    return out_file



r"""
Provide a convenience function for a Gaussian on top of a linear distribution.

Parameters
----------
area1
    The area of the Gaussian distribution
mu, sigma
    The location and scale of the first Gaussian
area2
    The area of the linear distributions respectively
m
    The slope of the linear distribution
b
    The y intercept of the linear distribution
x_lo
    The lower bound on which to normalize the linear distribution
x_hi
    The upper bound on which to normalize the linear distribution


Returns
-------
gauss_on_linear
    An instance of SumDists and rv_continuous, has methods of pdf, cdf, etc.
"""

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.linear import linear
from pygama.math.functions.sum_dists import SumDists

(linear_x_lo, linear_x_hi, n_sig, mu, sigma, n_bkg, m, b) = range(
    8
)  # the sum_dist array we will pass will be x_lo, x_hi, n_sig, mu, sigma, n_bkg, m, b
par_array = [(gaussian, [mu, sigma]), (linear, [linear_x_lo, linear_x_hi, m, b])]

gauss_on_linear = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "n_bkg", "m", "b"],
    name="gauss_on_linear",
)
